from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.ROI_time_bin_calculator import ROITimebinCalculator
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, DropDownMenu
from simba.utils.checks import check_float, check_int
from simba.utils.enums import ConfigKey, Formats, Keys, Links


class ROIAnalysisTimeBinsPopUp(ConfigReader, PopUpMixin):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        PopUpMixin.__init__(self, title="ROI ANALYSIS: TIME-BINS", size=(400, 400))
        self.animal_cnt_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SELECT NUMBER OF ANIMALS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ROI_DATA_ANALYSIS.value,
        )
        self.animal_cnt_dropdown = DropDownMenu(
            self.animal_cnt_frm,
            "# of animals",
            list(range(1, self.animal_cnt + 1)),
            labelwidth=20,
        )
        self.animal_cnt_dropdown.setChoices(1)
        self.animal_cnt_confirm_btn = Button(
            self.animal_cnt_frm,
            text="Confirm",
            command=lambda: self.create_settings_frm(),
        )
        self.animal_cnt_frm.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_dropdown.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_confirm_btn.grid(row=0, column=1, sticky=NW)
        self.main_frm.mainloop()

    def create_settings_frm(self):
        if hasattr(self, "setting_frm"):
            self.setting_frm.destroy()
            self.body_part_frm.destroy()
        self.setting_frm = LabelFrame(
            self.main_frm, text="SETTINGS", font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        self.choose_bp_frm(parent=self.setting_frm, bp_options=self.body_parts_lst)
        self.choose_bp_threshold_frm(parent=self.setting_frm)
        self.setting_frm.grid(row=1, column=0, sticky=NW)
        self.create_time_bin_entry()
        self.create_run_frm(run_function=self.run)

    def run(self):
        check_int(name="Time bin", value=self.time_bin_entrybox.entry_get, min_value=1)
        check_float(
            name="Probability threshold",
            value=self.probability_entry.entry_get,
            min_value=0.00,
            max_value=1.00,
        )
        self.config.set(
            ConfigKey.ROI_SETTINGS.value,
            ConfigKey.ROI_ANIMAL_CNT.value,
            str(self.animal_cnt_dropdown.getChoices()),
        )
        body_parts = []
        for cnt, dropdown in self.body_parts_dropdowns.items():
            self.config.set(
                ConfigKey.PROCESS_MOVEMENT_SETTINGS.value,
                "animal_{}_bp".format(str(cnt + 1)),
                str(dropdown.getChoices()),
            )
            body_parts.append(dropdown.getChoices())
        self.update_config()
        roi_time_bin_calculator = ROITimebinCalculator(
            config_path=self.config_path,
            bin_length=int(self.time_bin_entrybox.entry_get),
            threshold=float(self.probability_entry.entry_get),
            body_parts=body_parts,
        )
        roi_time_bin_calculator.run()
        roi_time_bin_calculator.save()
        self.root.destroy()


# ROIAnalysisTimeBinsPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/locomotion/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import glob
import os
import sys
from tkinter import *

from PIL import Image, ImageTk

import simba
from simba.labelling.extract_labelled_frames import AnnotationFrameExtractor
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.frame_mergerer_ffmpeg import FrameMergererFFmpeg
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon,
                                        CreateToolTip, DropDownMenu, Entry_Box,
                                        FileSelect, FolderSelect)
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists,
                                check_if_filepath_list_is_empty, check_int)
from simba.utils.data import convert_roi_definitions
from simba.utils.enums import Dtypes, Formats, Keys, Links, Options, Paths
from simba.utils.errors import (CountError, FrameRangeError, MixedMosaicError,
                                NoChoosenClassifierError, NoFilesFoundError,
                                NotDirectoryError)
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    get_video_meta_data)
from simba.video_processors.extract_seqframes import extract_seq_frames
from simba.video_processors.multi_cropper import MultiCropper
from simba.video_processors.px_to_mm import get_coordinates_nilsson
from simba.video_processors.video_processing import (
    VideoRotator, batch_convert_video_format, batch_create_frames,
    change_fps_of_multiple_videos, change_img_format, change_single_video_fps,
    clahe_enhance_video, clip_video_in_range, convert_to_mp4,
    convert_video_powerpoint_compatible_format, copy_img_folder,
    crop_multiple_videos, crop_single_video, downsample_video,
    extract_frame_range, extract_frames_single_video, frames_to_movie,
    gif_creator, multi_split_video, remove_beginning_of_video,
    superimpose_frame_count, video_concatenator, video_to_greyscale)

sys.setrecursionlimit(10**7)


class CLAHEPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="CLAHE VIDEO CONVERSION")
        clahe_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="Contrast Limited Adaptive Histogram Equalization",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        selected_video = FileSelect(
            clahe_frm,
            "Video path ",
            title="Select a video file",
            file_types=[("VIDEO", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        button_clahe = Button(
            clahe_frm,
            text="Apply CLAHE",
            command=lambda: clahe_enhance_video(file_path=selected_video.file_path),
        )
        clahe_frm.grid(row=0, sticky=W)
        selected_video.grid(row=0, sticky=W)
        button_clahe.grid(row=1, pady=5)
        # self.main_frm.mainloop()


# _ = CLAHEPopUp()


class CropVideoPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="CROP SINGLE VIDEO")
        crop_video_lbl_frm = LabelFrame(
            self.main_frm,
            text="Crop Video",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        selected_video = FileSelect(
            crop_video_lbl_frm,
            "Video path",
            title="Select a video file",
            lblwidth=20,
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        use_gpu_var_single = BooleanVar(value=False)
        use_gpu_cb_single = Checkbutton(
            crop_video_lbl_frm,
            text="Use GPU (reduced runtime)",
            variable=use_gpu_var_single,
        )
        button_crop_video_single = Button(
            crop_video_lbl_frm,
            text="Crop Video",
            command=lambda: crop_single_video(
                file_path=selected_video.file_path, gpu=use_gpu_var_single.get()
            ),
        )

        crop_video_lbl_frm_multiple = LabelFrame(
            self.main_frm,
            text="Fixed coordinates crop for multiple videos",
            font="bold",
            padx=5,
            pady=5,
        )
        input_folder = FolderSelect(
            crop_video_lbl_frm_multiple,
            "Video directory:",
            title="Select Folder with videos",
            lblwidth=20,
        )
        output_folder = FolderSelect(
            crop_video_lbl_frm_multiple,
            "Output directory:",
            title="Select a folder for your output videos",
            lblwidth=20,
        )
        use_gpu_var_multiple = BooleanVar(value=False)
        use_gpu_cb_multiple = Checkbutton(
            crop_video_lbl_frm_multiple,
            text="Use GPU (reduced runtime)",
            variable=use_gpu_var_multiple,
        )
        button_crop_video_multiple = Button(
            crop_video_lbl_frm_multiple,
            text="Crop Videos",
            command=lambda: crop_multiple_videos(
                directory_path=input_folder.folder_path,
                output_path=output_folder.folder_path,
                gpu=use_gpu_var_multiple.get(),
            ),
        )
        crop_video_lbl_frm.grid(row=0, sticky=NW)
        selected_video.grid(row=0, sticky=NW)
        use_gpu_cb_single.grid(row=1, column=0, sticky=NW)
        button_crop_video_single.grid(row=2, sticky=NW)
        crop_video_lbl_frm_multiple.grid(row=1, sticky=NW)
        input_folder.grid(row=0, sticky=NW)
        output_folder.grid(row=1, sticky=NW)
        use_gpu_cb_multiple.grid(row=2, sticky=NW)
        button_crop_video_multiple.grid(row=3, sticky=NW)


#         self.main_frm.mainloop()
#
# _ = CropVideoPopUp()


class ClipVideoPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="CLIP VIDEO")
        selected_video_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="Video path",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        selected_video = FileSelect(
            selected_video_frm,
            "FILE PATH: ",
            file_types=[("VIDEO", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        selected_video.grid(row=0, column=0, sticky="NW")
        use_gpu_frm = LabelFrame(self.main_frm, text="GPU", font="bold", padx=5, pady=5)
        self.use_gpu_var = BooleanVar(value=False)
        self.use_gpu_cb = Checkbutton(
            use_gpu_frm, text="Use GPU (reduced runtime)", variable=self.use_gpu_var
        )
        self.use_gpu_cb.grid(row=0, column=0, sticky=NW)
        method_1_frm = LabelFrame(
            self.main_frm, text="Method 1", font="bold", padx=5, pady=5
        )
        label_set_time_1 = Label(
            method_1_frm, text="Please enter the time frame in HH:MM:SS format"
        )
        start_time = Entry_Box(method_1_frm, "Start at (s):", "8")
        end_time = Entry_Box(method_1_frm, "End at (s):", "8")
        CreateToolTip(
            method_1_frm,
            "Method 1 will retrieve the specified time input. (eg: input of Start at: 00:00:00, End at: 00:01:00, will create a new video from the chosen video from the very start till it reaches the first minute of the video)",
        )
        method_2_frm = LabelFrame(
            self.main_frm, text="Method 2", font="bold", padx=5, pady=5
        )
        method_2_time = Entry_Box(method_2_frm, "Seconds:", "8", validation="numeric")
        label_method_2 = Label(
            method_2_frm,
            text="Method 2 will retrieve from the end of the video (e.g.,: an input of 3 seconds will get rid of the first 3 seconds of the video).",
        )
        button_cutvideo_method_1 = Button(
            method_1_frm,
            text="Cut Video",
            command=lambda: clip_video_in_range(
                file_path=selected_video.file_path,
                start_time=start_time.entry_get,
                end_time=end_time.entry_get,
                gpu=self.use_gpu_var.get(),
            ),
        )
        button_cutvideo_method_2 = Button(
            method_2_frm,
            text="Cut Video",
            command=lambda: remove_beginning_of_video(
                file_path=selected_video.file_path,
                time=method_2_time.entry_get,
                gpu=self.use_gpu_var.get(),
            ),
        )
        selected_video_frm.grid(row=0, sticky=NW)
        use_gpu_frm.grid(row=1, column=0, sticky=NW)
        method_1_frm.grid(row=2, sticky=NW, pady=5)
        label_set_time_1.grid(row=0, sticky=NW)
        start_time.grid(row=1, sticky=NW)
        end_time.grid(row=2, sticky=NW)
        button_cutvideo_method_1.grid(row=3, sticky=NW)
        method_2_frm.grid(row=3, sticky=NW, pady=5)
        label_method_2.grid(row=0, sticky=NW)
        method_2_time.grid(row=2, sticky=NW)
        button_cutvideo_method_2.grid(row=3, sticky=NW)


#       self.main_frm.mainloop()
# _ = ClipVideoPopUp()


class GreyscaleSingleVideoPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="GREYSCALE SINGLE VIDEO")
        settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="GREYSCALE VIDEO",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        self.selected_video = FileSelect(
            settings_frm,
            "VIDEO PATH",
            title="Select a video file",
            lblwidth=10,
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        self.use_gpu_var = BooleanVar(value=False)
        use_gpu_cb = Checkbutton(
            settings_frm, text="Use GPU (reduced runtime)", variable=self.use_gpu_var
        )

        settings_frm.grid(row=0, column=0, sticky="NW")
        self.selected_video.grid(row=0, column=0, sticky="NW")
        use_gpu_cb.grid(row=1, column=0, sticky="NW")
        self.create_run_frm(run_function=self.run)

    def run(self):
        check_file_exist_and_readable(file_path=self.selected_video.file_path)
        video_to_greyscale(
            file_path=self.selected_video.file_path, gpu=self.use_gpu_var.get()
        )


class SuperImposeFrameCountPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="SUPERIMPOSE FRAME COUNT ON VIDEO")
        settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SUPERIMPOSE FRAME COUNT ON VIDEO",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        self.selected_video = FileSelect(
            settings_frm,
            "VIDEO PATH",
            title="Select a video file",
            lblwidth=10,
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        self.use_gpu_var = BooleanVar(value=False)
        use_gpu_cb = Checkbutton(
            settings_frm, text="Use GPU (reduced runtime)", variable=self.use_gpu_var
        )

        settings_frm.grid(row=0, column=0, sticky="NW")
        self.selected_video.grid(row=0, column=0, sticky="NW")
        use_gpu_cb.grid(row=1, column=0, sticky="NW")
        self.create_run_frm(run_function=self.run)

    def run(self):
        check_file_exist_and_readable(file_path=self.selected_video.file_path)
        superimpose_frame_count(
            file_path=self.selected_video.file_path, gpu=self.use_gpu_var.get()
        )


class MultiShortenPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="CLIP VIDEO INTO MULTIPLE VIDEOS", size=(800, 200))
        settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="Split videos into different parts",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        self.selected_video = FileSelect(
            settings_frm,
            "Video path",
            title="Select a video file",
            lblwidth=10,
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        self.clip_cnt = Entry_Box(
            settings_frm, "# of clips", "10", validation="numeric"
        )
        self.use_gpu_var = BooleanVar(value=False)
        use_gpu_cb = Checkbutton(
            settings_frm, text="Use GPU (reduced runtime)", variable=self.use_gpu_var
        )
        confirm_settings_btn = Button(
            settings_frm, text="Confirm", command=lambda: self.show_start_stop()
        )
        settings_frm.grid(row=0, sticky=NW)
        self.selected_video.grid(row=1, sticky=NW, columnspan=2)
        self.clip_cnt.grid(row=2, sticky=NW)
        confirm_settings_btn.grid(row=2, column=1, sticky=W)
        use_gpu_cb.grid(row=3, column=0, sticky=W)
        instructions = Label(
            settings_frm,
            text="Enter clip start and stop times in HH:MM:SS format",
            fg="navy",
        )
        instructions.grid(row=4, column=0)

        batch_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="Batch change time",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        self.batch_start_entry = Entry_Box(batch_frm, "START:", "10")
        self.batch_start_entry.entry_set("00:00:00")
        self.batch_end_entry = Entry_Box(batch_frm, "END", "10")
        self.batch_end_entry.entry_set("00:00:00")
        batch_start_apply = Button(
            batch_frm, text="APPLY", command=lambda: self.batch_change(value="start")
        )
        batch_end_apply = Button(
            batch_frm, text="APPLY", command=lambda: self.batch_change(value="end")
        )

        batch_frm.grid(row=0, column=1, sticky=NW)
        self.batch_start_entry.grid(row=0, column=0, sticky=NW)
        batch_start_apply.grid(row=0, column=1, sticky=NW)
        self.batch_end_entry.grid(row=1, column=0, sticky=NW)
        batch_end_apply.grid(row=1, column=1, sticky=NW)

        # self.main_frm.mainloop()

    def show_start_stop(self):
        check_int(name="Number of clips", value=self.clip_cnt.entry_get)
        if hasattr(self, "table"):
            self.table.destroy()
        self.table = LabelFrame(self.main_frm)
        self.table.grid(row=2, column=0, sticky=NW)
        Label(self.table, text="Clip #").grid(row=0, column=0)
        Label(self.table, text="Start Time").grid(row=0, column=1, sticky=NW)
        Label(self.table, text="Stop Time").grid(row=0, column=2, sticky=NW)
        self.clip_names, self.start_times, self.end_times = [], [], []
        for i in range(int(self.clip_cnt.entry_get)):
            Label(self.table, text="Clip " + str(i + 1)).grid(row=i + 2, sticky=W)
            self.start_times.append(Entry(self.table))
            self.start_times[i].grid(row=i + 2, column=1, sticky=NW)
            self.end_times.append(Entry(self.table))
            self.end_times[i].grid(row=i + 2, column=2, sticky=NW)
        run_button = Button(
            self.table,
            text="Clip video",
            command=lambda: self.run_clipping(),
            fg="navy",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        run_button.grid(row=int(self.clip_cnt.entry_get) + 2, column=2, sticky=W)

    def batch_change(self, value: str):
        if not hasattr(self, "table"):
            raise CountError(
                msg="Select the number of video clippings first",
                source=self.__class__.__name__,
            )
        for start_time_entry, end_time_entry in zip(self.start_times, self.end_times):
            if value == "start":
                start_time_entry.delete(0, END)
                start_time_entry.insert(0, self.batch_start_entry.entry_get)
            else:
                end_time_entry.delete(0, END)
                end_time_entry.insert(0, self.batch_end_entry.entry_get)

    def run_clipping(self):
        start_times, end_times = [], []
        check_file_exist_and_readable(self.selected_video.file_path)
        for start_time, end_time in zip(self.start_times, self.end_times):
            start_times.append(start_time.get())
            end_times.append(end_time.get())
        multi_split_video(
            file_path=self.selected_video.file_path,
            start_times=start_times,
            end_times=end_times,
            gpu=self.use_gpu_var.get(),
        )


# _ = MultiShortenPopUp()


class ChangeImageFormatPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="CHANGE IMAGE FORMAT")

        self.input_folder_selected = FolderSelect(
            self.main_frm, "Image directory", title="Select folder with images:"
        )
        set_input_format_frm = LabelFrame(
            self.main_frm,
            text="Original image format",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            padx=15,
            pady=5,
        )
        set_output_format_frm = LabelFrame(
            self.main_frm,
            text="Output image format",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            padx=15,
            pady=5,
        )

        self.input_file_type, self.out_file_type = StringVar(), StringVar()
        input_png_rb = Radiobutton(
            set_input_format_frm,
            text=".png",
            variable=self.input_file_type,
            value="png",
        )
        input_jpeg_rb = Radiobutton(
            set_input_format_frm,
            text=".jpg",
            variable=self.input_file_type,
            value="jpg",
        )
        input_bmp_rb = Radiobutton(
            set_input_format_frm,
            text=".bmp",
            variable=self.input_file_type,
            value="bmp",
        )
        output_png_rb = Radiobutton(
            set_output_format_frm, text=".png", variable=self.out_file_type, value="png"
        )
        output_jpeg_rb = Radiobutton(
            set_output_format_frm, text=".jpg", variable=self.out_file_type, value="jpg"
        )
        output_bmp_rb = Radiobutton(
            set_output_format_frm, text=".bmp", variable=self.out_file_type, value="bmp"
        )
        run_btn = Button(
            self.main_frm,
            text="Convert image file format",
            command=lambda: self.run_img_conversion(),
        )
        self.input_folder_selected.grid(row=0, column=0)
        set_input_format_frm.grid(row=1, column=0, pady=5)
        set_output_format_frm.grid(row=2, column=0, pady=5)
        input_png_rb.grid(row=0, column=0)
        input_jpeg_rb.grid(row=1, column=0)
        input_bmp_rb.grid(row=2, column=0)
        output_png_rb.grid(row=0, column=0)
        output_jpeg_rb.grid(row=1, column=0)
        output_bmp_rb.grid(row=2, column=0)
        run_btn.grid(row=3, pady=5)

    def run_img_conversion(self):
        if len(os.listdir(self.input_folder_selected.folder_path)) == 0:
            raise NoFilesFoundError(
                msg="SIMBA ERROR: The input folder {} contains ZERO files.".format(
                    self.input_folder_selected.folder_path
                ),
                source=self.__class__.__name__,
            )
        change_img_format(
            directory=self.input_folder_selected.folder_path,
            file_type_in=self.input_file_type.get(),
            file_type_out=self.out_file_type.get(),
        )


class ConvertVideoPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, title="CONVERT VIDEO FORMAT", size=(200, 200))
        convert_multiple_videos_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="Convert multiple videos",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        video_dir = FolderSelect(
            convert_multiple_videos_frm,
            "Video directory",
            title="Select folder with videos",
            lblwidth=15,
        )
        video_format_options = ["mp4", "avi", "mov", "flv", "m4v"]
        original_format_dropdown = DropDownMenu(
            convert_multiple_videos_frm,
            "Input format",
            video_format_options,
            labelwidth=15,
        )
        output_format_dropdown = DropDownMenu(
            convert_multiple_videos_frm,
            "Input format",
            video_format_options,
            labelwidth=15,
        )
        original_format_dropdown.setChoices("avi")
        output_format_dropdown.setChoices("mp4")
        gpu_multiple_var = BooleanVar(value=False)
        gpu_multiple_cb = Checkbutton(
            convert_multiple_videos_frm,
            text="Use GPU (reduced runtime)",
            variable=gpu_multiple_var,
        )
        convert_multiple_btn = Button(
            convert_multiple_videos_frm,
            text="Convert multiple videos",
            command=lambda: batch_convert_video_format(
                directory=video_dir.folder_path,
                input_format=original_format_dropdown.getChoices(),
                output_format=output_format_dropdown.getChoices(),
                gpu=gpu_multiple_var.get(),
            ),
        )
        convert_single_video_frm = LabelFrame(
            self.main_frm,
            text="Convert single video",
            font=("Helvetica", 12, "bold"),
            padx=5,
            pady=5,
        )
        self.selected_video = FileSelect(
            convert_single_video_frm,
            "Video path",
            title="Select a video file",
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        self.output_format = StringVar()
        checkbox_v1 = Radiobutton(
            convert_single_video_frm,
            text="Convert to .mp4",
            variable=self.output_format,
            value="mp4",
        )
        checkbox_v2 = Radiobutton(
            convert_single_video_frm,
            text="Convert mp4 into PowerPoint supported format",
            variable=self.output_format,
            value="pptx",
        )
        self.gpu_single_var = BooleanVar(value=False)
        gpu_single_cb = Checkbutton(
            convert_single_video_frm,
            text="Use GPU (reduced runtime)",
            variable=self.gpu_single_var,
        )
        convert_single_btn = Button(
            convert_single_video_frm,
            text="Convert video format",
            command=lambda: self.convert_single(),
        )

        convert_multiple_videos_frm.grid(row=0, sticky=NW)
        video_dir.grid(row=0, sticky=NW)
        original_format_dropdown.grid(row=1, sticky=NW)
        output_format_dropdown.grid(row=2, sticky=NW)
        gpu_multiple_cb.grid(row=3, sticky=NW)
        convert_multiple_btn.grid(row=4, pady=10, sticky=NW)
        convert_single_video_frm.grid(row=1, sticky=NW)
        self.selected_video.grid(row=0, sticky=NW)
        checkbox_v1.grid(row=1, column=0, sticky=NW)
        checkbox_v2.grid(row=2, column=0, sticky=NW)
        gpu_single_cb.grid(row=3, column=0, sticky=NW)
        convert_single_btn.grid(row=4, column=0, pady=10, sticky=NW)

    def convert_single(self):
        if self.output_format.get() == "mp4":
            convert_to_mp4(
                file_path=self.selected_video.file_path, gpu=self.gpu_single_var.get()
            )
        if self.output_format.get() == "pptx":
            convert_video_powerpoint_compatible_format(
                file_path=self.selected_video.file_path, gpu=self.gpu_single_var.get()
            )


class ExtractSpecificFramesPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, title="EXTRACT DEFINED FRAMES", size=(200, 200))
        self.video_file_selected = FileSelect(
            self.main_frm,
            "Video path",
            title="Select a video file",
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        select_frames_frm = LabelFrame(
            self.main_frm, text="Frames to be extracted", padx=5, pady=5
        )
        self.start_frm = Entry_Box(select_frames_frm, "Start Frame:", "10")
        self.end_frm = Entry_Box(select_frames_frm, "End Frame:", "10")
        run_btn = Button(
            select_frames_frm,
            text="Extract Frames",
            command=lambda: self.start_frm_extraction(),
        )

        self.video_file_selected.grid(row=0, column=0, sticky=NW, pady=10)
        select_frames_frm.grid(row=1, column=0, sticky=NW)
        self.start_frm.grid(row=2, column=0, sticky=NW)
        self.end_frm.grid(row=3, column=0, sticky=NW)
        run_btn.grid(row=4, pady=5, sticky=NW)

    def start_frm_extraction(self):
        start_frame = self.start_frm.entry_get
        end_frame = self.end_frm.entry_get
        check_int(name="Start frame", value=start_frame)
        check_int(name="End frame", value=end_frame)
        if int(end_frame) < int(start_frame):
            raise FrameRangeError(
                msg="SIMBA ERROR: The end frame ({}) cannot come before the start frame ({})".format(
                    str(end_frame), str(start_frame)
                ),
                source=self.__class__.__name__,
            )
        video_meta_data = get_video_meta_data(
            video_path=self.video_file_selected.file_path
        )
        if int(start_frame) > video_meta_data["frame_count"]:
            raise FrameRangeError(
                msg="SIMBA ERROR: The start frame ({}) is larger than the number of frames in the video ({})".format(
                    str(start_frame), str(video_meta_data["frame_count"])
                ),
                source=self.__class__.__name__,
            )
        if int(end_frame) > video_meta_data["frame_count"]:
            raise FrameRangeError(
                msg="SIMBA ERROR: The end frame ({}) is larger than the number of frames in the video ({})".format(
                    str(end_frame), str(video_meta_data["frame_count"])
                ),
                source=self.__class__.__name__,
            )
        extract_frame_range(
            file_path=self.video_file_selected.file_path,
            start_frame=int(start_frame),
            end_frame=int(end_frame),
        )


class ExtractAllFramesPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, title="EXTRACT ALL FRAMES", size=(200, 200))
        single_video_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="Single video",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        video_path = FileSelect(
            single_video_frm,
            "Video path",
            title="Select a video file",
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        single_video_btn = Button(
            single_video_frm,
            text="Extract Frames (Single video)",
            command=lambda: extract_frames_single_video(file_path=video_path.file_path),
        )
        multiple_videos_frm = LabelFrame(
            self.main_frm, text="Multiple videos", padx=5, pady=5, font="bold"
        )
        folder_path = FolderSelect(
            multiple_videos_frm, "Folder path", title=" Select video folder"
        )
        multiple_video_btn = Button(
            multiple_videos_frm,
            text="Extract Frames (Multiple videos)",
            command=lambda: batch_create_frames(directory=folder_path.folder_path),
        )
        single_video_frm.grid(row=0, sticky=NW, pady=10)
        video_path.grid(row=0, sticky=NW)
        single_video_btn.grid(row=1, sticky=W, pady=10)
        multiple_videos_frm.grid(row=1, sticky=W, pady=10)
        folder_path.grid(row=0, sticky=W)
        multiple_video_btn.grid(row=1, sticky=W, pady=10)


class MultiCropPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, title="MULTI-CROP", size=(500, 300))
        self.input_folder = FolderSelect(
            self.main_frm, "Input Video Folder", lblwidth=15
        )
        self.output_folder = FolderSelect(self.main_frm, "Output Folder", lblwidth=15)
        video_options = ["mp4", "avi", "mov", "flv", "m4v"]
        self.video_type_dropdown = DropDownMenu(
            self.main_frm, "Video type:", video_options, "15"
        )
        self.video_type_dropdown.setChoices("mp4")
        self.crop_cnt_dropdown = DropDownMenu(
            self.main_frm, "Crop count:", list(range(1, 31)), "15"
        )
        self.crop_cnt_dropdown.setChoices(2)
        self.use_gpu_var = BooleanVar(value=False)
        self.use_gpu_cb = Checkbutton(
            self.main_frm, text="Use GPU (reduced runtime)", variable=self.use_gpu_var
        )
        self.create_run_frm(run_function=self.run)
        self.input_folder.grid(row=0, sticky=NW)
        self.output_folder.grid(row=1, sticky=NW)
        self.video_type_dropdown.grid(row=2, sticky=NW)
        self.crop_cnt_dropdown.grid(row=3, sticky=NW)
        self.use_gpu_cb.grid(row=4, sticky=NW)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        check_if_dir_exists(in_dir=self.input_folder.folder_path)
        check_if_dir_exists(in_dir=self.output_folder.folder_path)
        MultiCropper(
            file_type=self.video_type_dropdown.getChoices(),
            input_folder=self.input_folder.folder_path,
            output_folder=self.output_folder.folder_path,
            crop_cnt=self.crop_cnt_dropdown.getChoices(),
            gpu=self.use_gpu_var.get(),
        ).run()


# MultiCropPopUp()


class ChangeFpsSingleVideoPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(
            self, title="CHANGE FRAME RATE: SINGLE VIDEO", size=(200, 200)
        )
        video_path = FileSelect(
            self.main_frm,
            "Video path",
            title="Select a video file",
            lblwidth=10,
            file_types=[("VIDEO", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        fps_entry_box = Entry_Box(
            self.main_frm, "Output FPS:", "10", validation="numeric"
        )
        gpu_var = BooleanVar(value=False)
        gpu_cb = Checkbutton(
            self.main_frm, text="Use GPU (reduced runtime)", variable=gpu_var
        )
        run_btn = Button(
            self.main_frm,
            text="Convert",
            command=lambda: change_single_video_fps(
                file_path=video_path.file_path,
                fps=fps_entry_box.entry_get,
                gpu=gpu_var.get(),
            ),
        )
        video_path.grid(row=0, sticky=NW)
        fps_entry_box.grid(row=1, sticky=NW)
        gpu_cb.grid(row=2, sticky=NW)
        run_btn.grid(row=3, sticky=NW)


class ChangeFpsMultipleVideosPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(
            self, title="CHANGE FRAME RATE: MULTIPLE VIDEO", size=(400, 200)
        )
        folder_path = FolderSelect(
            self.main_frm,
            "Folder path",
            title="Select folder with videos: ",
            lblwidth="10",
        )
        fps_entry = Entry_Box(self.main_frm, "Output FPS: ", "10", validation="numeric")
        gpu_var = BooleanVar(value=False)
        gpu_cb = Checkbutton(
            self.main_frm, text="Use GPU (reduced runtime)", variable=gpu_var
        )
        run_btn = Button(
            self.main_frm,
            text="Convert",
            command=lambda: change_fps_of_multiple_videos(
                directory=folder_path.folder_path,
                fps=fps_entry.entry_get,
                gpu=gpu_var.get(),
            ),
        )
        folder_path.grid(row=0, sticky=NW)
        fps_entry.grid(row=1, sticky=NW)
        gpu_cb.grid(row=2, sticky=NW)
        run_btn.grid(row=3, sticky=NW)


class ExtractSEQFramesPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(
            self, title="EXTRACT ALL FRAMES FROM SEQ FILE", size=(200, 200)
        )
        video_path = FileSelect(
            self.main_frm,
            "Video Path",
            title="Select a video file: ",
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        run_btn = Button(
            self.main_frm,
            text="Extract All Frames",
            command=lambda: extract_seq_frames(video_path.file_path),
        )
        video_path.grid(row=0)
        run_btn.grid(row=1)


class MergeFrames2VideoPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, title="MERGE IMAGES TO VIDEO", size=(250, 250))
        self.folder_path = FolderSelect(
            self.main_frm, "IMAGE DIRECTORY", title="Select directory with frames: "
        )
        settings_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            padx=5,
            pady=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.img_format_entry_box = Entry_Box(
            settings_frm, "IMAGE FORMAT (e.g. png): ", "20"
        )
        self.bitrate_entry_box = Entry_Box(
            settings_frm, "BITRATE (e.g. 8000): ", "20", validation="numeric"
        )
        self.fps_entry = Entry_Box(settings_frm, "FPS: ", "20", validation="numeric")
        self.gpu_var = BooleanVar(value=False)
        gpu_cb = Checkbutton(
            settings_frm, text="Use GPU (decreased runtime)", variable=self.gpu_var
        )
        run_btn = Button(settings_frm, text="Create Video", command=lambda: self.run())
        settings_frm.grid(row=1, pady=10)
        self.folder_path.grid(row=0, column=0, pady=10)
        self.img_format_entry_box.grid(row=1, column=0, sticky=W)
        self.fps_entry.grid(row=2, column=0, sticky=W, pady=5)
        self.bitrate_entry_box.grid(row=3, column=0, sticky=W, pady=5)
        gpu_cb.grid(row=4, column=0, sticky=W, pady=5)
        run_btn.grid(row=5, column=1, sticky=E, pady=10)

    def run(self):
        img_format = self.img_format_entry_box.entry_get
        bitrate = self.bitrate_entry_box.entry_get
        fps = self.fps_entry.entry_get
        _ = frames_to_movie(
            directory=self.folder_path.folder_path,
            fps=fps,
            bitrate=bitrate,
            img_format=img_format,
            gpu=self.gpu_var.get(),
        )


class CreateGIFPopUP(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, title="CREATE GIF FROM VIDEO", size=(250, 250))
        settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        selected_video = FileSelect(
            settings_frm,
            "Video path: ",
            title="Select a video file",
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        start_time_entry_box = Entry_Box(
            settings_frm, "Start time (s): ", "16", validation="numeric"
        )
        duration_entry_box = Entry_Box(
            settings_frm, "Duration (s): ", "16", validation="numeric"
        )
        width_entry_box = Entry_Box(settings_frm, "Width: ", "16", validation="numeric")
        gpu_var = BooleanVar()
        gpu_cb = Checkbutton(
            settings_frm, text="Use GPU (decreased runtime)", variable=gpu_var
        )
        width_instructions_1 = Label(
            settings_frm,
            text="example Width: 240, 360, 480, 720, 1080",
            font=("Times", 10, "italic"),
        )
        width_instructions_2 = Label(
            settings_frm,
            text="Aspect ratio is kept (i.e., height is automatically computed)",
            font=("Times", 10, "italic"),
        )
        run_btn = Button(
            settings_frm,
            text="CREATE GIF",
            command=lambda: gif_creator(
                file_path=selected_video.file_path,
                start_time=start_time_entry_box.entry_get,
                duration=duration_entry_box.entry_get,
                width=width_entry_box.entry_get,
                gpu=gpu_var.get(),
            ),
        )
        settings_frm.grid(row=0, sticky=NW)
        selected_video.grid(row=0, sticky=NW, pady=5)
        start_time_entry_box.grid(row=1, sticky=NW)
        duration_entry_box.grid(row=2, sticky=NW)
        width_entry_box.grid(row=3, sticky=NW)
        gpu_cb.grid(row=4, column=0, sticky=NW)
        width_instructions_1.grid(row=4, sticky=NW)
        width_instructions_2.grid(row=5, sticky=NW)
        run_btn.grid(row=6, sticky=NW, pady=10)


class CalculatePixelsPerMMInVideoPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(
            self, title="CALCULATE PIXELS PER MILLIMETER IN VIDEO", size=(200, 200)
        )
        self.video_path = FileSelect(
            self.main_frm,
            "Select a video file: ",
            title="Select a video file",
            file_types=[("VIDEO", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        self.known_distance = Entry_Box(
            self.main_frm, "Known length in real life (mm): ", "0", validation="numeric"
        )
        run_btn = Button(
            self.main_frm, text="GET PIXELS PER MILLIMETER", command=lambda: self.run()
        )
        self.video_path.grid(row=0, column=0, pady=10, sticky=W)
        self.known_distance.grid(row=1, column=0, pady=10, sticky=W)
        run_btn.grid(row=2, column=0, pady=10)

    def run(self):
        check_file_exist_and_readable(file_path=self.video_path.file_path)
        check_int(name="Distance", value=self.known_distance.entry_get, min_value=1)
        _ = get_video_meta_data(video_path=self.video_path.file_path)
        mm_cnt = get_coordinates_nilsson(
            self.video_path.file_path, self.known_distance.entry_get
        )
        print(
            f"1 PIXEL REPRESENTS {round(mm_cnt, 4)} MILLIMETERS IN VIDEO {os.path.basename(self.video_path.file_path)}."
        )


class ConcatenatingVideosPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, title="CONCATENATE VIDEOS", size=(300, 300))
        settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        video_path_1 = FileSelect(
            settings_frm,
            "First video path: ",
            title="Select a video file",
            lblwidth=15,
            file_types=[("VIDEO", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        video_path_2 = FileSelect(
            settings_frm,
            "Second video path: ",
            title="Select a video file",
            lblwidth=15,
            file_types=[("VIDEO", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        resolutions = ["Video 1", "Video 2", 320, 640, 720, 1280, 1980]
        resolution_dropdown = DropDownMenu(
            settings_frm, "Resolution:", resolutions, "15"
        )
        resolution_dropdown.setChoices(resolutions[0])
        use_gpu_var = BooleanVar(value=False)
        horizontal = BooleanVar(value=False)
        horizontal_radio_btn = Radiobutton(
            settings_frm,
            text="Horizontal concatenation",
            variable=horizontal,
            value=True,
        )
        use_gpu_cb = Checkbutton(
            settings_frm, text="Use GPU (reduced runtime)", variable=use_gpu_var
        )
        vertical_radio_btn = Radiobutton(
            settings_frm,
            text="Vertical concatenation",
            variable=horizontal,
            value=False,
        )
        run_btn = Button(
            self.main_frm,
            text="RUN",
            font=("Helvetica", 12, "bold"),
            command=lambda: video_concatenator(
                video_one_path=video_path_1.file_path,
                video_two_path=video_path_2.file_path,
                resolution=resolution_dropdown.getChoices(),
                horizontal=horizontal.get(),
                gpu=use_gpu_var.get(),
            ),
        )

        settings_frm.grid(row=0, column=0, sticky=NW)
        video_path_1.grid(row=0, column=0, sticky=NW)
        video_path_2.grid(row=1, column=0, sticky=NW)
        resolution_dropdown.grid(row=2, column=0, sticky=NW)
        use_gpu_cb.grid(row=3, column=0, sticky=NW)
        horizontal_radio_btn.grid(row=4, column=0, sticky=NW)
        vertical_radio_btn.grid(row=5, column=0, sticky=NW)
        run_btn.grid(row=1, column=0, sticky=NW)


class ConcatenatorPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str or None):
        PopUpMixin.__init__(self, title="MERGE (CONCATENATE) VIDEOS")
        self.config_path = config_path
        self.select_video_cnt_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="VIDEOS #",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.CONCAT_VIDEOS.value,
        )
        self.select_video_cnt_dropdown = DropDownMenu(
            self.select_video_cnt_frm, "VIDEOS #", list(range(2, 21)), "15"
        )
        self.select_video_cnt_dropdown.setChoices(2)
        self.select_video_cnt_btn = Button(
            self.select_video_cnt_frm,
            text="SELECT",
            command=lambda: self.populate_table(),
        )
        self.select_video_cnt_frm.grid(row=0, column=0, sticky=NW)
        self.select_video_cnt_dropdown.grid(row=0, column=0, sticky=NW)
        self.select_video_cnt_btn.grid(row=0, column=1, sticky=NW)

    def populate_table(self):
        if hasattr(self, "video_table_frm"):
            self.video_table_frm.destroy()
            self.join_type_frm.destroy()
        self.video_table_frm = LabelFrame(
            self.main_frm,
            text="VIDEO PATHS",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.video_table_frm.grid(row=1, sticky=NW)
        self.join_type_frm = LabelFrame(
            self.main_frm,
            text="JOIN TYPE",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.join_type_frm.grid(row=2, sticky=NW)
        self.videos_dict = {}
        for cnt in range(int(self.select_video_cnt_dropdown.getChoices())):
            self.videos_dict[cnt] = FileSelect(
                self.video_table_frm,
                "Video {}: ".format(str(cnt + 1)),
                title="Select a video file",
                file_types=[("VIDEO", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
            )
            self.videos_dict[cnt].grid(row=cnt, column=0, sticky=NW)

        self.join_type_var = StringVar()
        self.icons_dict = {}
        simba_dir = os.path.dirname(simba.__file__)
        icon_assets_dir = os.path.join(simba_dir, Paths.ICON_ASSETS.value)
        concat_icon_dir = os.path.join(icon_assets_dir, "concat_icons")
        for file_cnt, file_path in enumerate(glob.glob(concat_icon_dir + "/*")):
            _, file_name, _ = get_fn_ext(file_path)
            self.icons_dict[file_name] = {}
            self.icons_dict[file_name]["img"] = ImageTk.PhotoImage(
                Image.open(file_path)
            )
            self.icons_dict[file_name]["btn"] = Radiobutton(
                self.join_type_frm,
                text=file_name,
                variable=self.join_type_var,
                value=file_name,
            )
            self.icons_dict[file_name]["btn"].config(
                image=self.icons_dict[file_name]["img"]
            )
            self.icons_dict[file_name]["btn"].image = self.icons_dict[file_name]["img"]
            self.icons_dict[file_name]["btn"].grid(row=0, column=file_cnt, sticky=NW)
        self.join_type_var.set(value="mosaic")
        self.resolution_frm = LabelFrame(
            self.main_frm,
            text="RESOLUTION",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.resolution_width = DropDownMenu(
            self.resolution_frm, "Width", ["480", "640", "1280", "1920", "2560"], "15"
        )
        self.resolution_width.setChoices("640")
        self.resolution_height = DropDownMenu(
            self.resolution_frm, "Height", ["480", "640", "1280", "1920", "2560"], "15"
        )
        self.resolution_height.setChoices("480")
        self.gpu_frm = LabelFrame(
            self.main_frm,
            text="GPU",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.use_gpu_var = BooleanVar(value=False)
        use_gpu_cb = Checkbutton(
            self.gpu_frm, text="Use GPU (reduced runtime)", variable=self.use_gpu_var
        )
        use_gpu_cb.grid(row=0, column=0, sticky="NW")
        self.resolution_frm.grid(row=3, column=0, sticky=NW)
        self.gpu_frm.grid(row=4, column=0, sticky="NW")
        self.resolution_width.grid(row=0, column=0, sticky=NW)
        self.resolution_height.grid(row=1, column=0, sticky=NW)

        run_btn = Button(self.main_frm, text="RUN", command=lambda: self.run())
        run_btn.grid(row=5, column=0, sticky=NW)

    def run(self):
        videos_info = {}
        for cnt, (video_name, video_data) in enumerate(self.videos_dict.items()):
            _ = get_video_meta_data(video_path=video_data.file_path)
            videos_info["Video {}".format(str(cnt + 1))] = video_data.file_path

        if (len(videos_info.keys()) < 3) & (self.join_type_var.get() == "mixed_mosaic"):
            raise MixedMosaicError(
                msg="Ff using the mixed mosaic join type, please tick check-boxes for at least three video types.",
                source=self.__class__.__name__,
            )
        if (len(videos_info.keys()) < 3) & (self.join_type_var.get() == "mosaic"):
            self.join_type_var.set(value="vertical")

        _ = FrameMergererFFmpeg(
            config_path=self.config_path,
            frame_types=videos_info,
            video_height=int(self.resolution_height.getChoices()),
            video_width=int(self.resolution_width.getChoices()),
            concat_type=self.join_type_var.get(),
            gpu=self.use_gpu_var.get(),
        )


class VideoRotatorPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="ROTATE VIDEOS")
        self.save_dir_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SAVE LOCATION",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        self.save_dir = FolderSelect(self.save_dir_frm, "Save directory:", lblwidth=20)

        self.setting_frm = LabelFrame(
            self.main_frm, text="SETTINGS", font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        self.use_gpu_var = BooleanVar(value=False)
        self.use_ffmpeg_var = BooleanVar(value=False)
        use_gpu_cb = Checkbutton(
            self.setting_frm,
            text="Use GPU (reduced runtime)",
            variable=self.use_gpu_var,
        )
        use_ffmpeg_cb = Checkbutton(
            self.setting_frm,
            text="Use FFMpeg (reduced runtime over default OpenCV)",
            variable=self.use_ffmpeg_var,
        )
        use_gpu_cb.grid(row=0, column=0, sticky=NW)
        use_ffmpeg_cb.grid(row=1, column=0, sticky=NW)

        self.rotate_dir_frm = LabelFrame(
            self.main_frm,
            text="ROTATE VIDEOS IN DIRECTORY",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.input_dir = FolderSelect(
            self.rotate_dir_frm, "Video directory:", lblwidth=20
        )
        self.run_dir = Button(
            self.rotate_dir_frm,
            text="RUN",
            fg="blue",
            command=lambda: self.run(
                input_path=self.input_dir.folder_path,
                output_path=self.save_dir.folder_path,
            ),
        )

        self.rotate_video_frm = LabelFrame(
            self.main_frm,
            text="ROTATE SINGLE VIDEO",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.input_file = FileSelect(
            self.rotate_video_frm,
            "Video path:",
            lblwidth=20,
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        self.run_file = Button(
            self.rotate_video_frm,
            text="RUN",
            fg="blue",
            command=lambda: self.run(
                input_path=self.input_file.file_path,
                output_path=self.save_dir.folder_path,
            ),
        )

        self.save_dir_frm.grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=0, column=0, sticky=NW)

        self.setting_frm.grid(row=1, column=0, sticky=NW)

        self.rotate_dir_frm.grid(row=2, column=0, sticky=NW)
        self.input_dir.grid(row=0, column=0, sticky=NW)
        self.run_dir.grid(row=1, column=0, sticky=NW)

        self.rotate_video_frm.grid(row=3, column=0, sticky=NW)
        self.input_file.grid(row=0, column=0, sticky=NW)
        self.run_file.grid(row=1, column=0, sticky=NW)

    def run(self, input_path: str, output_path: str):
        check_if_dir_exists(in_dir=output_path)
        rotator = VideoRotator(
            input_path=input_path,
            output_dir=output_path,
            ffmpeg=self.use_ffmpeg_var.get(),
            gpu=self.use_gpu_var.get(),
        )
        rotator.run()


class VideoTemporalJoinPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="TEMPORAL JOIN VIDEOS")
        self.settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        self.input_dir = FolderSelect(
            self.settings_frm, "INPUT DIRECTORY:", lblwidth=20
        )
        self.file_format = DropDownMenu(
            self.settings_frm, "INPUT FORMAT:", Options.VIDEO_FORMAT_OPTIONS.value, "20"
        )
        self.use_gpu_var = BooleanVar(value=False)
        use_gpu_cb = Checkbutton(
            self.settings_frm,
            text="Use GPU (reduced runtime)",
            variable=self.use_gpu_var,
        )
        self.file_format.setChoices(Options.VIDEO_FORMAT_OPTIONS.value[0])
        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.input_dir.grid(row=0, column=0, sticky=NW)
        self.file_format.grid(row=1, column=0, sticky=NW)
        use_gpu_cb.grid(row=2, column=0, sticky="NW")
        self.create_run_frm(run_function=self.run)

    def run(self):
        check_if_dir_exists(in_dir=self.input_dir.folder_path)
        print(f"Concatenating videos in {self.input_dir.folder_path} directory...")
        save_path = os.path.join(self.input_dir.folder_path, f"concatenated.mp4")
        concatenate_videos_in_folder(
            in_folder=self.input_dir.folder_path,
            save_path=save_path,
            remove_splits=False,
            video_format=self.file_format.getChoices(),
            gpu=self.use_gpu_var.get(),
        )


class ImportFrameDirectoryPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="IMPORT FRAME DIRECTORY")
        ConfigReader.__init__(self, config_path=config_path)
        self.frame_folder = FolderSelect(
            self.main_frm,
            "FRAME DIRECTORY:",
            title="Select the main directory with frame folders",
        )
        import_btn = Button(
            self.main_frm, text="IMPORT FRAMES", fg="blue", command=lambda: self.run()
        )

        self.frame_folder.grid(row=0, column=0, sticky=NW)
        import_btn.grid(row=1, column=0, sticky=NW)

    def run(self):
        if not os.path.isdir(self.frame_folder.folder_path):
            raise NotDirectoryError(
                msg=f"SIMBA ERROR: {self.frame_folder.folder_path} is not a valid directory.",
                source=self.__class__.__name__,
            )
        copy_img_folder(
            config_path=self.config_path, source=self.frame_folder.folder_path
        )


class ExtractAnnotationFramesPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(
            self, config_path=config_path, title="EXTRACT ANNOTATED FRAMES"
        )
        ConfigReader.__init__(self, config_path=config_path)
        self.create_clf_checkboxes(main_frm=self.main_frm, clfs=self.clf_names)
        self.settings_frm = LabelFrame(
            self.main_frm,
            text="STYLE SETTINGS",
            font=("Helvetica", 12, "bold"),
            pady=5,
            padx=5,
        )
        down_sample_resolution_options = ["None", "2x", "3x", "4x", "5x"]
        self.resolution_downsample_dropdown = DropDownMenu(
            self.settings_frm,
            "Down-sample images:",
            down_sample_resolution_options,
            "25",
        )
        self.resolution_downsample_dropdown.setChoices(
            down_sample_resolution_options[0]
        )
        self.settings_frm.grid(row=self.children_cnt_main(), column=0, sticky=NW)
        self.resolution_downsample_dropdown.grid(row=0, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        check_if_filepath_list_is_empty(
            self.target_file_paths,
            error_msg=f"SIMBA ERROR: Zero files found in the {self.targets_folder} directory",
        )
        downsample_setting = self.resolution_downsample_dropdown.getChoices()
        if downsample_setting != Dtypes.NONE.value:
            downsample_setting = int("".join(filter(str.isdigit, downsample_setting)))
        clfs = []
        for clf_name, selection in self.clf_selections.items():
            if selection.get():
                clfs.append(clf_name)
        if len(clfs) == 0:
            raise NoChoosenClassifierError(source=self.__class__.__name__)
        settings = {"downsample": downsample_setting}

        frame_extractor = AnnotationFrameExtractor(
            config_path=self.config_path, clfs=clfs, settings=settings
        )
        frame_extractor.run()


class DownsampleVideoPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="DOWN-SAMPLE VIDEO RESOLUTION")
        instructions = Label(
            self.main_frm,
            text="Choose only one of the following method (Custom or Default)",
        )
        choose_video_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SELECT VIDEO",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.DOWNSAMPLE.value,
        )
        self.video_path_selected = FileSelect(
            choose_video_frm,
            "Video path",
            title="Select a video file",
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        gpu_frm = LabelFrame(
            self.main_frm,
            text="GPU",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
            padx=5,
            pady=5,
        )
        self.use_gpu_var = BooleanVar(value=False)
        use_gpu_cb = Checkbutton(
            gpu_frm, text="Use GPU (reduced runtime)", variable=self.use_gpu_var
        )
        use_gpu_cb.grid(row=0, column=0, sticky="NW")
        custom_frm = LabelFrame(
            self.main_frm,
            text="Custom resolution",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
            padx=5,
            pady=5,
        )
        self.entry_width = Entry_Box(custom_frm, "Width", "10", validation="numeric")
        self.entry_height = Entry_Box(custom_frm, "Height", "10", validation="numeric")

        self.custom_downsample_btn = Button(
            custom_frm,
            text="Downsample to custom resolution",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
            command=lambda: self.custom_downsample(),
        )
        default_frm = LabelFrame(
            self.main_frm, text="Default resolution", font="bold", padx=5, pady=5
        )
        self.radio_btns = {}
        self.var = StringVar()
        for custom_cnt, resolution_radiobtn in enumerate(self.resolutions):
            self.radio_btns[resolution_radiobtn] = Radiobutton(
                default_frm,
                text=resolution_radiobtn,
                variable=self.var,
                value=resolution_radiobtn,
            )
            self.radio_btns[resolution_radiobtn].grid(row=custom_cnt, sticky=NW)

        self.default_downsample_btn = Button(
            default_frm,
            text="Downsample to default resolution",
            command=lambda: self.default_downsample(),
        )
        instructions.grid(row=0, sticky=NW, pady=10)
        choose_video_frm.grid(row=1, column=0, sticky=NW)
        gpu_frm.grid(row=2, column=0, sticky=NW)
        self.video_path_selected.grid(row=0, column=0, sticky=NW)
        custom_frm.grid(row=3, column=0, sticky=NW)
        self.entry_width.grid(row=0, column=0, sticky=NW)
        self.entry_height.grid(row=1, column=0, sticky=NW)
        self.custom_downsample_btn.grid(row=3, column=0, sticky=NW)
        default_frm.grid(row=5, column=0, sticky=NW)
        self.default_downsample_btn.grid(
            row=len(self.resolutions) + 1, column=0, sticky=NW
        )

    def custom_downsample(self):
        width = self.entry_width.entry_get
        height = self.entry_height.entry_get
        check_int(name="Video width", value=width)
        check_int(name="Video height", value=height)
        check_file_exist_and_readable(self.video_path_selected.file_path)
        downsample_video(
            file_path=self.video_path_selected.file_path,
            video_width=int(width),
            video_height=int(height),
            gpu=self.use_gpu_var.get(),
        )

    def default_downsample(self):
        resolution = self.var.get()
        width, height = (
            resolution.split("×", 2)[0].strip(),
            resolution.split("×", 2)[1].strip(),
        )
        check_file_exist_and_readable(self.video_path_selected.file_path)
        downsample_video(
            file_path=self.video_path_selected.file_path,
            video_width=int(width),
            video_height=int(height),
            gpu=self.use_gpu_var.get(),
        )


class ConvertROIDefinitionsPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="CONVERT ROI DEFINITIONS")
        settings_frm = LabelFrame(
            self.main_frm, text="SETTINGS", font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        self.roi_definitions_file_select = FileSelect(
            settings_frm,
            "ROI DEFINITIONS PATH (H5)",
            title="SELECT H5 FILE",
            lblwidth=20,
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
        )
        self.save_dir = FolderSelect(
            settings_frm, "SAVE DIRECTORY", title="SELECT H5 FILE", lblwidth=20
        )
        settings_frm.grid(row=0, column=0, sticky=NW)
        self.roi_definitions_file_select.grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=1, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

    def run(self):
        check_file_exist_and_readable(
            file_path=self.roi_definitions_file_select.file_path
        )
        check_if_dir_exists(in_dir=self.save_dir.folder_path)
        _ = convert_roi_definitions(
            roi_definitions_path=self.roi_definitions_file_select.file_path,
            save_dir=self.save_dir.folder_path,
        )



import glob
import itertools
import math
from copy import deepcopy
from itertools import combinations

import cv2
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from numba import jit, prange
from scipy.spatial import ConvexHull


class EgocentricalAlignmentFeaturizer(object):
    def __init__(
        self,
        data_path: str,
        anchor: str = "Centroid",
        fps: int = 30,
        visualize: bool = False,
    ):
        self.data_files = glob.glob(data_path + "/*.csv")
        self.anchor = (f"{anchor}_x", f"{anchor}_y")
        self.fps = fps
        self.visualize = visualize
        self.img_size = (500, 500)
        self.rolling_window_sizes = {}
        for i in [1, 1.5, 3]:
            self.rolling_window_sizes[f"{str(i)}s"] = int(fps * i)
        for i in [2, 4, 10]:
            self.rolling_window_sizes[f"{str(1/i)}s"] = int(fps / i)
        self.run()

    def run(self):
        for file_path in self.data_files:
            df = pd.read_csv(file_path, header=[0, 1, 2], index_col=0)
            df.columns = df.columns.droplevel().map("_".join)
            self.bp_headers, self.bp_dict = {}, {}
            for i, j in zip(["_x", "_y", "_p"], ["x", "y", "p"]):
                self.bp_headers[j] = [x for x in df.columns if x.endswith(i)]
            df = df[self.bp_headers["x"] + self.bp_headers["y"]]
            self.scaled_df = deepcopy(df)
            for bp in self.bp_headers["x"]:
                self.bp_dict[bp.rstrip("_x")] = (bp, bp.rstrip("_x") + "_y")
            df["correction_x"] = df[self.anchor[0]] - (self.img_size[0] / 2)
            df["correction_y"] = df[self.anchor[1]] - (self.img_size[1] / 2)
            for c in self.bp_dict.values():
                self.scaled_df[c[0]] = self.scaled_df[c[0]] - df["correction_x"]
                self.scaled_df[c[1]] = self.scaled_df[c[1]] - df["correction_y"]
            self.scaled_df = self.scaled_df.fillna((self.img_size[0] / 2))
            if self.visualize:
                self.visualize()
            self.featurize()

    def visualize(self):
        max_x, max_y = np.nanmax(
            self.scaled_df[self.bp_headers["x"]].values
        ), np.nanmax(self.scaled_df[self.bp_headers["y"]].values)
        img = np.zeros(shape=[int(max_x), int(max_y), 3], dtype=np.uint8)
        for frm in range(len(self.scaled_df)):
            frm_data = self.scaled_df.iloc[frm].astype(int)
            frm_img = deepcopy(img)
            for bp_name, bp in self.bp_dict.items():
                x, y = frm_data[bp[0]], frm_data[bp[1]]
                cv2.circle(frm_img, (int(x), int(y)), 0, (255, 255, 0), 8)
            for bp_c in combinations(list(self.bp_dict.keys()), 2):
                bp_1_x, bp_1_y = self.bp_dict[bp_c[0]][0], self.bp_dict[bp_c[0]][1]
                bp_2_x, bp_2_y = self.bp_dict[bp_c[1]][0], self.bp_dict[bp_c[1]][1]
                point_one, point_two = (frm_data[bp_1_x], frm_data[bp_1_y]), (
                    frm_data[bp_2_x],
                    frm_data[bp_2_y],
                )
                cv2.line(frm_img, point_one, point_two, (255, 255, 0), 1)
            cv2.imshow("img", frm_img)
            cv2.waitKey(33)

    @staticmethod
    @jit(nopython=True, cache=True, fastmath=True)
    def three_point_angles(data: np.array):
        results = np.full((data.shape[0]), 0)
        for i in prange(data.shape[0]):
            angle = math.degrees(
                math.atan2(data[i][5] - data[i][3], data[i][4] - data[i][2])
                - math.atan2(data[i][1] - data[i][3], data[i][0] - data[i][2])
            )
            if angle < 0:
                angle += 360
            results[i] = angle
        return results

    @staticmethod
    def subhull_calculator(data: np.array):
        results = np.full((len(data)), np.nan)
        data = np.reshape(data.values, (len(data), -1, 2))
        for cnt, i in enumerate(data):
            results[cnt] = ConvexHull(i).area
        return results.astype(int)

    @staticmethod
    @jit(nopython=True)
    def euclidean_distance(bp_1_x_vals, bp_2_x_vals, bp_1_y_vals, bp_2_y_vals):
        return np.sqrt(
            (bp_1_x_vals - bp_2_x_vals) ** 2 + (bp_1_y_vals - bp_2_y_vals) ** 2
        )

    @staticmethod
    def convex_hull_calculator_mp(data: np.array):
        results = np.full((data.shape[0]), np.nan)
        data = np.reshape(data.values, (len(data), -1, 2))
        for cnt, i in enumerate(data):
            results[cnt] = ConvexHull(i).area
        return results.astype(int)

    def featurize(self):
        three_point_combinations = np.array(
            list(combinations(list(self.bp_dict.keys()), 3))
        )
        four_point_combinations = np.array(
            list(combinations(list(self.bp_dict.keys()), 4))
        )
        two_point_combinations = np.array(
            list(combinations(list(self.bp_dict.keys()), 2))
        )
        results = pd.DataFrame()
        split_data = np.array_split(self.scaled_df, 100)
        hull_area = Parallel(n_jobs=-1, verbose=0, backend="threading")(
            delayed(self.convex_hull_calculator_mp)(x) for x in split_data
        )
        results["hull_area"] = np.concatenate(hull_area).ravel().tolist()
        for c in three_point_combinations:
            col_names = list(sum([(x + "_x", y + "_y") for (x, y) in zip(c, c)], ()))
            split_data = np.array_split(self.scaled_df[col_names], 100)
            three_point_hull = Parallel(n_jobs=-1, verbose=0, backend="threading")(
                delayed(self.subhull_calculator)(x) for x in split_data
            )
            results[f"hull_{c[0]}_{c[1]}_{c[2]}"] = (
                np.concatenate(three_point_hull).ravel().tolist()
            )
            results[f"angle_{c[0]}_{c[1]}_{c[2]}"] = self.three_point_angles(
                data=self.scaled_df[col_names].values
            )
        for c in four_point_combinations:
            col_names = list(sum([(x + "_x", y + "_y") for (x, y) in zip(c, c)], ()))
            split_data = np.array_split(self.scaled_df[col_names], 100)
            four_point_hull = Parallel(n_jobs=-1, verbose=0, backend="threading")(
                delayed(self.subhull_calculator)(x) for x in split_data
            )
            results[f"hull_{c[0]}_{c[1]}_{c[2]}_{c[3]}"] = (
                np.concatenate(four_point_hull).ravel().tolist()
            )
        for c in two_point_combinations:
            col_names = list(sum([(x + "_x", y + "_y") for (x, y) in zip(c, c)], ()))
            results[f"distance_{c[0]}_{c[1]}"] = self.euclidean_distance(
                self.scaled_df[col_names[0]].values,
                self.scaled_df[col_names[2]].values,
                self.scaled_df[col_names[1]].values,
                self.scaled_df[col_names[3]].values,
            )

        for c, t in list(
            itertools.product(results.columns, self.rolling_window_sizes.keys())
        ):
            results[f"{c}_rolling_{t}_window_mean"] = (
                results[c]
                .rolling(int(self.rolling_window_sizes[t]), min_periods=1)
                .mean()
            )
            results[f"{c}_rolling_{t}_window_stdev"] = (
                results[c]
                .rolling(int(self.rolling_window_sizes[t]), min_periods=1)
                .std()
            )
        self.results = results.fillna(-1)


aligner = EgocentricalAlignmentFeaturizer(
    data_path="/Users/simon/Desktop/envs/simba_dev/simba/features_scripts/misc/test_data_mouse_OF",
    visualize=False,
)



import pytest
import pandas as pd
from simba.utils.read_write import read_df
from simba.utils.data import (detect_bouts,
                              plug_holes_shortest_bout,
                              create_color_palettes,
                              create_color_palette)

@pytest.mark.parametrize("data_path, target_lst, fps", [('tests/data/test_projects/two_c57/project_folder/csv/machine_results/Together_1.csv', ['Attack', 'Sniffing'], 30)])
def test_detect_bouts(data_path, target_lst, fps):
    data_df = read_df(file_path=data_path, file_type='csv')
    results = detect_bouts(data_df=data_df, target_lst=target_lst, fps=fps)

def test_plug_holes_shortest_bout():
    data_df = pd.DataFrame(data=[1, 0, 1, 1, 1], columns=['target'])
    results = plug_holes_shortest_bout(data_df=data_df, clf_name='target', fps=10, shortest_bout=2000)
    pd.testing.assert_frame_equal(results, pd.DataFrame(data=[1, 1, 1, 1, 1], columns=['target']))

def test_create_color_palettes():
    results = create_color_palettes(no_animals=2, map_size=2)
    assert results == [[[255.0, 0.0, 255.0], [0.0, 255.0, 255.0]], [[102.0, 127.5, 0.0], [102.0, 255.0, 255.0]]]

def test_create_color_palette():
    results = create_color_palette(pallete_name='jet', increments=3)
    assert results == [[127.5, 0.0, 0.0], [255.0, 212.5, 0.0], [0.0, 229.81481481481478, 255.0], [0.0, 0.0, 127.5]]
    results = create_color_palette(pallete_name='jet', increments=3, as_rgb_ratio=True)
    assert results == [[0.5, 0.0, 0.0], [1.0, 0.8333333333333334, 0.0], [0.0, 0.9012345679012345, 1.0], [0.0, 0.0, 0.5]]
    results = create_color_palette(pallete_name='jet', increments=3, as_hex=True)
    assert results == ['#800000', '#ffd400', '#00e6ff', '#000080']



import os

import cv2

from simba.enums import Paths
from simba.misc_tools import get_fn_ext
from simba.utils.errors import InvalidVideoFileError


def get_video_meta_data(video_path: str):
    """
    Helper to read video metadata (fps, resolution, frame cnt etc.) from video file.

    Parameters
    ----------
    video_path: str
        Path to video file.

    Returns
    -------
    vdata: dict
        Python dictionary holding video meta data

    Notes
    ----------

    Examples
    >>> get_video_meta_data('/Users/simon/Desktop/envs/simba_dev/tests/test_data/video_tests/Together_1.avi')
    {'video_name': 'Together_1', 'fps': 30, 'width': 400, 'height': 600, 'frame_count': 300, 'resolution_str': '400 x 600', 'video_length_s': 10}

    """
    d = os.path.join(
        os.path.dirname(os.path.abspath("__file__")),
        "tests/test_data/video_tests/Together_1.avi",
    )
    video_data = {}
    cap = cv2.VideoCapture(video_path)
    _, video_data["video_name"], _ = get_fn_ext(video_path)
    video_data["fps"] = int(cap.get(cv2.CAP_PROP_FPS))
    video_data["width"] = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    video_data["height"] = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    video_data["frame_count"] = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    for k, v in video_data.items():
        if v == 0:
            raise InvalidVideoFileError(
                msg=f'SIMBA WARNING: Video {video_data["video_name"]} has {k} of {str(v)}.'
            )
    video_data["resolution_str"] = str(
        f'{video_data["width"]} x {video_data["height"]}'
    )
    video_data["video_length_s"] = int(video_data["frame_count"] / video_data["fps"])
    return video_data


# = get_video_meta_data('tests/test_data/video_tests/Together_1.avi')



import glob
import os
from copy import deepcopy

import pandas as pd

DATA_DIR = "/Users/simon/Desktop/envs/simba_dev/tests/test_data/visualization_tests/project_folder/csv/machine_results"
SAVE_DIR = "/Users/simon/Desktop/envs/simba_dev/tests/test_data/visualization_tests/project_folder/csv/output"
FIRST_CLF = "Swimming_normal"
SECOND_CLF = "Swimming_fast"


data_paths = glob.glob(DATA_DIR + "/*.csv")
for file_path in data_paths:
    df = pd.read_csv(file_path, index_col=0)
    output_df = deepcopy(df)
    missing_cols = list(set(["FIRST_CLF", "SECOND_CLF"]) - set(list(df.columns)))
    if missing_cols:
        print(f"ERROR: {missing_cols} column(s) missing in file {file_path}")
    df.loc[df.FIRST_CLF == 1, SECOND_CLF] = 0



"""
SimBA (Simple Behavioral Analysis)
https://github.com/sgoldenlab/simba
Contributors.
https://github.com/sgoldenlab/simba#contributors-
Licensed under GNU Lesser General Public License v3.0
"""

import setuptools

with open("docs/project_description.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="Simba-UW-tf-dev",
    version="1.84.7",
    author="Simon Nilsson, Jia Jie Choong, Sophia Hwang",
    author_email="sronilsson@gmail.com",
    description="Toolkit for computer classification of complex social behaviors in experimental animals",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sgoldenlab/simba",
    install_requires=requirements,
    license='GNU Lesser General Public License v3 (LGPLv3)',
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
    ),
    entry_points={'console_scripts':['simba=simba.SimBA:main'],}
)



import pytest
from copy import deepcopy
import numpy as np
import os
import pandas as pd
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
import simba
from simba.utils.enums import Paths
from scipy.signal import savgol_filter


def test_euclidean_distance():
    x1, x2 = np.full((10, 1), 1), np.full((10, 1), 2)
    y1, y2 = np.full((10, 1), 2), np.full((10, 1), 1)
    results = FeatureExtractionMixin().euclidean_distance(bp_1_x=x1, bp_2_x=x2, bp_1_y=y1, bp_2_y=y2, px_per_mm=4.56)
    assert results.shape == x1.shape
    assert np.all(results == results[0])
    assert round(results[0][0], 2) == 0.31

def test_angle3pt():
    results = FeatureExtractionMixin.angle3pt(ax=122.0, ay=198.0, bx=237.0, by=138.0, cx=191.0, cy=109)
    assert round(results, 2) == 59.78

def test_angle3pt_serialized():
    x, y, z = np.full((6, 2), 1), np.full((6, 2), 2), np.full((6, 2), 3)
    coordinates = np.hstack((x, y, z))
    results = FeatureExtractionMixin.angle3pt_serialized(data=coordinates)
    assert np.all(results == results[0])
    assert results.shape[0] == x.shape[0]
    assert results[0] == 180

def test_convex_hull_calculator_mp():
    coordinates = np.array([[1, 2], [4, 7], [9, 1], [3, 6]])
    results = FeatureExtractionMixin.convex_hull_calculator_mp(arr=coordinates, px_per_mm=1)
    assert round(results, 2) == 21.76


def test_count_values_in_range():
    data = np.array([[0.01, 0.5, 0.9, 0.3, 0.6], [0.01, 0.5, 0.9, 0.3, 0.6]])
    results = FeatureExtractionMixin.count_values_in_range(data=data, ranges=np.array([[0.0, 0.25], [0.25, 0.5], [0.5, 1.0]]))
    assert results.shape == (2, 3)
    assert np.array_equal(results[0], [1, 2, 3])

def test_framewise_euclidean_distance_roi():
    loc_1 = np.array([[180, 1], [ 24, 35], [6, 129]]).astype('float32')
    loc_2 = np.array([[117, 134]]).astype('float32')
    results = FeatureExtractionMixin.framewise_euclidean_distance_roi(location_1=loc_1, location_2=loc_2, px_per_mm=4.56, centimeter=False).astype(int)
    assert np.array_equal(results, [32, 29, 24])

def test_framewise_inside_rectangle_roi():
    bp_loc = np.array([[5, 7], [4, 2]])
    roi_coords = np.array([[0, 0], [10, 9]])
    results = FeatureExtractionMixin.framewise_inside_rectangle_roi(bp_location=bp_loc, roi_coords=roi_coords)
    assert np.array_equal(results, [1, 1])
    roi_coords = np.array([[10, 0], [10, 9]])
    results = FeatureExtractionMixin.framewise_inside_rectangle_roi(bp_location=bp_loc, roi_coords=roi_coords)
    assert np.array_equal(results, [0, 0])

def test_framewise_inside_polygon_roi():
    bp_loc = np.array([[5, 7], [4, 2]])
    roi_coords = np.array([[5, 7], [6, 4],  [8, 5], [2, 7]]).astype(np.float32)
    results = FeatureExtractionMixin.framewise_inside_polygon_roi(bp_location=bp_loc, roi_coords=roi_coords)
    assert np.array_equal(results, [1, 0])
    
def test_windowed_frequentist_distribution_tests():
    data = np.random.randint(1, 10, size=(10))
    results = FeatureExtractionMixin.windowed_frequentist_distribution_tests(data=data, fps=25, feature_name='Anima_1_velocity')
    assert len(results.columns) == 4; assert len(results) == data.shape[0]
    assert results._is_numeric_mixed_type

def test_cdist():
    array_1 = np.array([[3, 9], [5, 1], [2, 5]]).astype('float32')
    array_2 = np.array([[1, 5], [3, 6], [4, 9]]).astype('float32')
    results = FeatureExtractionMixin.cdist(array_1=array_1, array_2=array_2).astype('int')
    assert np.array_equal(results, np.array([[4, 3, 1], [5, 5, 8], [1, 1, 4]]))

def test_create_shifted_df():
    df = pd.DataFrame(data=[10, 95, 85], columns=['Feature_1'])
    results = FeatureExtractionMixin.create_shifted_df(df=df)
    assert list(results.columns) == ['Feature_1', 'Feature_1_shifted']
    assert len(results) == len(df)
    assert results._is_numeric_mixed_type

@pytest.mark.parametrize("pose", ['2 animals 16 body-parts',
                                  '2 animals 14 body-parts',
                                  '2 animals 8 body-parts',
                                  '1 animal 8 body-parts',
                                  '1 animal 7 body-parts',
                                  '1 animal 4 body-parts',
                                  '1 animal 9 body-parts'])
def test_get_feature_extraction_headers(pose):
    simba_dir = os.path.dirname(simba.__file__)
    feature_categories_csv_path = os.path.join(simba_dir, Paths.SIMBA_FEATURE_EXTRACTION_COL_NAMES_PATH.value)
    bps = list(pd.read_csv(feature_categories_csv_path)[pose])
    assert type(bps) == list

def test_minimum_bounding_rectangle():
    points = np.array([[7, 6], [9, 4], [4, 7], [5,1]])
    results = FeatureExtractionMixin.minimum_bounding_rectangle(points=points).astype(int)
    assert np.array_equal(results, np.array([[9, 1], [4, 1], [4, 7], [8, 7]]))

def test_framewise_euclidean_distance():
    loc_1 = np.array([[108, 162]]).astype(np.float32)
    loc_2 = np.array([[91, 11]]).astype(np.float32)
    results_mm = FeatureExtractionMixin.framewise_euclidean_distance_roi(location_1=loc_1, location_2=loc_2, px_per_mm=4.56, centimeter=False).astype(int)
    results_cm = FeatureExtractionMixin.framewise_euclidean_distance_roi(location_1=loc_1, location_2=loc_2, px_per_mm=4.56, centimeter=True).astype(int)
    assert int(results_mm / 10) == results_cm

@pytest.mark.parametrize("time_window", [1, 3])
def test_dataframe_gaussian_smoother(time_window):
    in_df = pd.DataFrame(data=[10, 95, 85], columns=['Feature_1'])
    out_df = deepcopy(in_df)
    for c in in_df.columns: out_df[c] = out_df[c].rolling(window=int(time_window), win_type='gaussian', center=True).mean(std=5).fillna(out_df[c]).abs().astype(int)
    if time_window == 1: assert out_df.equals(in_df)
    else: assert out_df.equals(pd.DataFrame(data=[10, 63, 85], columns=['Feature_1']))

@pytest.mark.parametrize("time_window", [5])
def test_dataframe_savgol_smoother(time_window):
    in_df = pd.DataFrame(data=[10, 95, 85, 109, 205], columns=['Feature_1'])
    out_df = deepcopy(in_df)
    for c in in_df.columns:
        out_df[c] = savgol_filter(x=in_df[c].to_numpy(), window_length=time_window, polyorder=3, mode='nearest').astype(int)
    assert out_df.equals(pd.DataFrame(data=[32, 68, 92, 126, 182], columns=['Feature_1']))



#test_euclidean_distance()
# test_angle3pt()
# test_angle3pt_serialized()
# test_convex_hull_calculator_mp()
# test_count_values_in_range()
# test_framewise_euclidean_distance_roi()
# test_framewise_inside_rectangle_roi()
# test_framewise_inside_polygon_roi()



from typing import List

import numpy as np
from numba import jit, prange, typed
from scipy import stats


class FeatureExtractionCircularMixin(object):
    """
    Mixin for circular statistics. Support for multiple animals and base
    radial directions derived from two or three body-parts.

    .. important::
        See references below for  mature packages computing extensive circular measurements.

    .. image:: _static/img/circular_statistics.png
       :width: 800
       :align: center

    References
    ----------
    .. [1] `pycircstat <https://github.com/circstat/pycircstat>`_.
    .. [2] `circstat <https://www.mathworks.com/matlabcentral/fileexchange/10676-circular-statistics-toolbox-directional-statistics>`_.
    .. [3] `pingouin.circular <https://github.com/raphaelvallat/pingouin/blob/master/pingouin/circular.py>`_.
    .. [4] `pycircular <https://github.com/albahnsen/pycircular>`_.
    .. [5] `scipy.stats.directional_stats <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.directional_stats.html>`_.
    .. [6] `astropy.stats.circstats <https://docs.astropy.org/en/stable/_modules/astropy/stats/circstats.html>`_.
    """

    def __init__(self):
        pass

    @staticmethod
    def rolling_mean_dispersion(
        data: np.ndarray, time_windows: np.ndarray, fps: int
    ) -> np.ndarray:
        """
        Compute the angular mean dispersion (circular mean) in degrees within rolling temporal windows.

        :parameter np.ndarray data: 1d array with feature values in degrees.
        :parameter np.ndarray time_windows: Rolling time-windows as floats in seconds. E.g., [0.2, 0.4, 0.6]
        :parameter int fps: fps of the recorded video
        :returns np.ndarray: Size data.shape[0] x time_windows.shape[0] array

        .. image:: _static/img/mean_rolling_timeseries_angle.png
           :width: 1000
           :align: center

        .. attention::
           The returned values represents the angular mean dispersion in the time-window ``[current_frame-time_window->current_frame]``.
           `-1` is returned when ``current_frame-time_window`` is less than 0.

        :example:
        >>> data = np.random.normal(loc=45, scale=1, size=20)
        >>> FeatureExtractionCircularMixin().rolling_mean_dispersion(data=data,time_windows=np.array([0.5]), fps=10)
        >>> [ [-1],[-1],[-1],[-1], [-1],[44],[44],[43],[44],[44],[44],[44],[44],[44],[44],[45],[45],[45],[45],[45]])
        """

        results = np.full((data.shape[0], time_windows.shape[0]), -1)
        for time_window in prange(time_windows.shape[0]):
            jump_frms = int(time_windows[time_window] * fps)
            for current_frm in prange(jump_frms, results.shape[0] + 1):
                data_window = np.deg2rad(data[current_frm - jump_frms : current_frm])
                results[current_frm - 1][time_window] = np.rad2deg(
                    stats.circmean(data_window)
                ).astype(int)
        return results

    @staticmethod
    @jit(nopython=True)
    def degrees_to_compass_cardinal(degree_angles: np.ndarray) -> List[str]:
        """
        Convert degree angles to cardinal direction bucket e.g., 0 -> "N", 180 -> "S"

        .. note::
           To convert cardinal literals to integers, map using ``simba.utils.enums.lookups.cardinality_to_integer_lookup``.
           To convert integers to cardinal literals, map using ``simba.utils.enums.lookups.integer_to_cardinality_lookup``.

        :parameter degree_angles nose_loc: 1d array of degrees. Note: return by ``self.head_direction``.
        :return List[str]: List of strings representing frame-wise cardinality

        :example:
        >>> data = np.array(list(range(0, 405, 45)))
        >>> FeatureExtractionCircularMixin().degrees_to_compass_cardinal(degree_angles=data)
        >>> ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW', 'N']
        """
        results = typed.List(["str"])
        DIRECTIONS = ["N", "NE", "E", "SE", "S", "SW", "W", "NW"]
        for i in prange(degree_angles.shape[0]):
            ix = round(degree_angles[i] / (360.0 / len(DIRECTIONS)))
            direction = DIRECTIONS[ix % len(DIRECTIONS)]
            results.append(direction)
        return results[1:]

    @staticmethod
    @jit(nopython=True)
    def direction_three_bps(
        nose_loc: np.ndarray, left_ear_loc: np.ndarray, right_ear_loc: np.ndarray
    ) -> np.ndarray:
        """
        Jitted helper to compute the degree angle from three body-parts. Computes the angle in degrees left_ear <-> nose
        and right_ear_nose and returns the midpoint.

        .. image:: _static/img/angle_from_3_bps.png
          :width: 600
          :align: center

        :parameter ndarray nose_loc: 2D array of size len(frames)x2 representing nose coordinates
        :parameter ndarray left_ear_loc: 2D array of size len(frames)x2 representing left ear coordinates
        :parameter ndarray right_ear_loc: 2D array of size len(frames)x2 representing right ear coordinates
        :return np.ndarray: Array of size nose_loc.shape[0] with direction in degrees.

        :example:
        >>> nose_loc = np.random.randint(low=0, high=500, size=(50, 2)).astype('float32')
        >>> left_ear_loc = np.random.randint(low=0, high=500, size=(50, 2)).astype('float32')
        >>> right_ear_loc = np.random.randint(low=0, high=500, size=(50, 2)).astype('float32')
        >>> results = FeatureExtractionCircularMixin().direction_three_bps(nose_loc=nose_loc, left_ear_loc=left_ear_loc, right_ear_loc=right_ear_loc)
        """

        results = np.full((nose_loc.shape[0]), np.nan)
        for i in prange(nose_loc.shape[0]):
            left_ear_to_nose = np.degrees(
                np.arctan2(
                    left_ear_loc[i][0] - nose_loc[i][1],
                    left_ear_loc[i][1] - nose_loc[i][0],
                )
            )
            right_ear_nose = np.degrees(
                np.arctan2(
                    right_ear_loc[i][0] - nose_loc[i][1],
                    right_ear_loc[i][1] - nose_loc[i][0],
                )
            )
            results[i] = ((left_ear_to_nose + right_ear_nose) % 360) / 2
        return results

    @staticmethod
    @jit(nopython=True)
    def direction_two_bps(bp_x: np.ndarray, bp_y: np.ndarray) -> np.ndarray:
        """
        Jitted method computing degree directionality from two body-parts. E.g., ``nape`` and ``nose``,
        or ``swim_bladder`` and ``tail``.

        .. image:: _static/img/angle_from_2_bps.png
           :width: 1200
           :align: center

        :parameter np.ndarray bp_x: Size len(frames) x 2 representing x and y coordinates for first body-part.
        :parameter np.ndarray bp_y: Size len(frames) x 2 representing x and y coordinates for second body-part.
        :return np.ndarray: Frame-wise directionality in degrees.

        :example:
        >>> swim_bladder_loc = np.random.randint(low=0, high=500, size=(50, 2)).astype('float32')
        >>> tail_loc = np.random.randint(low=0, high=500, size=(50, 2)).astype('float32')
        >>> results = FeatureExtractionCircularMixin().direction_three_bps(bp_x=swim_bladder_loc, bp_y=tail_loc)
        """

        results = np.full((bp_x.shape[0]), np.nan)
        for i in prange(bp_x.shape[0]):
            angle_degrees = np.degrees(
                np.arctan2(bp_x[i][0] - bp_y[i][0], bp_y[i][1] - bp_x[i][1])
            )
            angle_degrees = angle_degrees + 360 if angle_degrees < 0 else angle_degrees
            results[i] = angle_degrees
        return results

    @staticmethod
    @jit(nopython=True)
    def rolling_resultant_vector_length(
        data: np.ndarray, fps: int, time_windows: np.ndarray = np.array([1.0])
    ) -> np.ndarray:
        """
        Jitted helper computing the mean resultant vector within rolling time window.

        .. note:
           Adapted from ``pingouin.circular.circ_r``.

        .. attention::
           The returned values represents resultant vector length in the time-window ``[current_frame-time_window->current_frame]``.
           `-1` is returned where ``current_frame-time_window`` is less than 0.

        :parameter np.ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Rolling time-window as float in seconds. Default: 1s rolling time-window.
        :parameter int fps: fps of the recorded video
        :returns np.ndarray: Size len(frames) representing resultant vector length in the prior ``time_window``.

        :example:
        >>> data_1, data_2 = np.random.normal(loc=45, scale=3, size=20), np.random.normal(loc=45, scale=150, size=30)
        >>> data = np.hstack([data_1, data_2])
        >>> FeatureExtractionCircularMixin().rolling_resultant_vector_length(data=data,time_windows=np.array([1]), fps=20)
        >>> [[-1],[-1],[-1],[-1],[-1],[ 0.998],[ 0.999],[ 0.999],[ 0.999],[ 0.997],[ 0.997],[ 0.996],[ 0.996],[ 0.996],[ 0.998],[ 0.998],[ 0.999],[ 0.998],[ 0.998],[ 0.999],[ 0.998],[ 0.655],[ 0.644],[ 0.367],[ 0.277],[ 0.193],[ 0.582],[ 0.200],[ 0.254],[ 0.235],[ 0.126],[ 0.145],[ 0.336],[ 0.719],[ 0.682],[ 0.780],[ 0.576],[ 0.314],[ 0.333],[ 0.291],[ 0.304],[ 0.095],[ 0.410],[ 0.061],[ 0.052],[ 0.262],[ 0.217],[ 0.485],[ 0.411],[ 0.736]])
        """

        data = np.deg2rad(data)
        results = np.full((data.shape[0], time_windows.shape[0]), -1.0)
        for time_window_cnt in prange(time_windows.shape[0]):
            window_size = int(time_windows[time_window_cnt] * fps)
            for window_end in prange(window_size, data.shape[0] + 1, 1):
                window_data = data[window_end - window_size : window_end]
                w = np.ones(window_data.shape[0])
                r = np.nansum(np.multiply(w, np.exp(1j * window_data)))
                results[window_end - 1][time_window_cnt] = np.abs(r) / np.nansum(w)
        return results

    @staticmethod
    @jit(nopython=True)
    def _helper_rayleigh_z(data: np.ndarray, window_size: int):
        results = np.full((data.shape[0], 2), np.nan)
        for i in range(data.shape[0]):
            r = window_size * data[i]
            results[i][0] = (r**2) / window_size
            results[i][1] = np.exp(
                np.sqrt(1 + 4 * window_size + 4 * (window_size**2 - r**2))
                - (1 + 2 * window_size)
            )
        return results

    def rolling_rayleigh_z(
        self, data: np.ndarray, fps: int, time_window: float = 1.0
    ) -> np.array:
        """
        Compute Rayleigh Z (test of non-uniformity) of circular data within rolling time-window.

        .. note:
           Adapted from ``pingouin.circular.circ_rayleigh``.

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Rolling time-window as float in seconds. Default: 1s rolling time-window.
        :parameter int fps: fps of the recorded video
        :returns np.ndarray: Size data.shape[0] x 2 with Rayleigh Z statistics in first column and associated p_values in second column
        """

        results, window_size = np.full((data.shape[0], 2), np.nan), int(
            time_window * fps
        )
        resultant_vector_lengths = (
            FeatureExtractionCircularMixin().rolling_resultant_vector_length(
                data=data, fps=fps, time_window=time_window
            )
        )
        return np.nan_to_num(
            self._helper_rayleigh_z(
                data=resultant_vector_lengths, window_size=window_size
            ),
            nan=-1.0,
        )

    @staticmethod
    @jit(nopython=True)
    def rolling_circular_correlation(
        data_x: np.ndarray, data_y: np.ndarray, fps: int, time_window: float = 1.0
    ) -> np.ndarray:
        """
        Compute correlations between two angular distributions in rolling time-windows.

        .. image:: _static/img/cicle_correlation.png
           :width: 800
           :align: center

        :parameter np.ndarray data_x: Angular data for e.g., Animal 1
        :parameter np.ndarray data_y: Angular data for e.g., Animal 2
        :parameter int fps: Frame-rate of video.
        :parameter float time_window: Time window in seconds.
        :return np.ndarray: Correlation coefficients for ``data_x`` and ``data_y``.
        """

        data_x, data_y = np.deg2rad(data_x), np.deg2rad(data_y)
        results = np.full((data_x.shape[0]), np.nan)
        window_size = int(time_window * fps)
        for window_start in prange(0, data_x.shape[0] - window_size + 1):
            data_x_window = data_x[window_start : window_start + window_size]
            data_y_window = data_y[window_start : window_start + window_size]
            x_sin = np.sin(
                data_x_window
                - np.angle(np.nansum(np.multiply(1, np.exp(1j * data_x_window))))
            )
            y_sin = np.sin(
                data_y_window
                - np.angle(np.nansum(np.multiply(1, np.exp(1j * data_y_window))))
            )
            r = np.sum(x_sin * y_sin) / np.sqrt(np.sum(x_sin**2) * np.sum(y_sin**2))
            results[window_start + window_size] = (
                np.sqrt(
                    (data_x_window.shape[0] * (x_sin**2).mean() * (y_sin**2).mean())
                    / np.mean(x_sin**2 * y_sin**2)
                )
                * r
            )

        return results

    @staticmethod
    def rolling_circular_stdev(
        data: np.ndarray, fps: int, time_windows: np.ndarray
    ) -> np.ndarray:
        """
        Compute standard deviation of angular data in rolling time windows.

        .. image:: _static/img/angle_stdev.png
           :width: 800
           :align: center

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Rolling time-window as float in seconds.
        :parameter int fps: fps of the recorded video
        :returns np.ndarray: Size data.shape[0] x time_windows.shape[0] with angular standard deviations in rolling time windows.

        :example:
        >>> data_1, data_2 = np.random.normal(loc=180, scale=3, size=10), np.random.normal(loc=90, scale=700, size=10)
        >>> data = np.hstack([data_1, data_2])
        >>> results = FeatureExtractionCircularMixin().rolling_circular_stdev(data=data, time_windows=np.array([1]), fps=10)
        """

        data = np.deg2rad(data)
        results = np.full((data.shape[0], time_windows.shape[0]), 0.0)
        for time_window_cnt in prange(time_windows.shape[0]):
            window_size = int(time_windows[time_window_cnt] * fps)
            for window_end in prange(window_size, data.shape[0] + 1, 1):
                window_data = data[window_end - window_size : window_end]
                results[window_end - 1][time_window_cnt] = stats.circvar(window_data)
        return np.round(results, 4)

    @staticmethod
    @jit(nopython=True)
    def rolling_angular_difference(
        data: np.ndarray, fps: int, time_windows: np.ndarray
    ):
        """
        Computes the angular difference in the current frame versus N seconds previously.
        For example, if the current angle is 45 degrees, and the angle N seconds previously was 350 degrees, then the difference
        is 55 degrees.

        .. note::
           Frames where current frame - N seconds prior equal a negative value is populated with 0.

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Rolling time-window as float in seconds.
        :parameter int fps: fps of the recorded video

        :example:
        >>> data = np.array([350, 350, 1, 1])
        >>> FeatureExtractionCircularMixin().rolling_angular_difference(data=data, fps=1, time_windows=np.array([1]))
        >>> [[ 0.], [ 0.], [11.], [ 0.]]
        """

        data = np.deg2rad(data)
        results = np.full((data.shape[0], time_windows.shape[0]), 0.0)
        for time_window_cnt in prange(time_windows.shape[0]):
            window_size = int(time_windows[time_window_cnt] * fps)
            for window_end in prange(window_size, data.shape[0], 1):
                point_one, point_two = data[window_end - window_size], data[window_end]
                print(point_one, point_two)
                distance = np.pi - np.abs(np.pi - np.abs(point_one - point_two))
                results[window_end][time_window_cnt] = np.rad2deg(distance)

        return results

    @staticmethod
    @jit(nopython=True)
    def agg_angular_difference_timebins(
        data: np.ndarray, fps: int, time_windows: np.ndarray
    ):
        """
        Compute the difference between the median angle in the current time-window versus the previous time window.
        For example, computes the difference between the mean angle in the first 1s of the video versus
        the second 1s of the video, the second 1s of the video versus the third 1s of the video, ... etc.

        .. note::
           The first time-bin of the video can't be compared against the prior time-bin of the video and the results
           for this first time-bin will be populated with `0`.

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Rolling time-window as float in seconds.
        :parameter int fps: fps of the recorded video

        :example:
        >>> data = np.random.normal(loc=45, scale=3, size=20)
        >>> FeatureExtractionCircularMixin().agg_angular_difference_timebins(data=data,time_windows=np.array([1]), fps=5)
        """

        data = np.deg2rad(data)
        results = np.full((data.shape[0], time_windows.shape[0]), 0.0)
        for time_window_cnt in prange(time_windows.shape[0]):
            window_size = int(time_windows[time_window_cnt] * fps)
            prior_window = [0, window_size]
            for win_cnt, window_end in enumerate(
                prange(int(window_size * 2), data.shape[0] + 1, window_size)
            ):
                window_start = (window_end - window_size) - 1
                current_data = data[window_start:window_end]
                prior_data = data[prior_window[0] : prior_window[1]]
                prior_median = np.arctan2(
                    np.median(np.cos(prior_data)), np.median(np.sin(prior_data))
                )
                current_median = np.arctan2(
                    np.median(np.cos(current_data)), np.median(np.sin(current_data))
                )
                distance = np.pi - np.abs(np.pi - np.abs(prior_median - current_median))
                results[window_start:window_end, win_cnt] = np.rad2deg(distance)
                prior_window = [window_start, window_end]

        return results

    @staticmethod
    @jit(nopython=True)
    def instantaneous_angular_velocity(data: np.ndarray, bin_size: int = 1):
        """
        Jitted compute of absolute angular change in the smallest possible time bin.

        .. note::
            If the smallest possible frame-to-frame time-bin in Video 1 is 33ms (recorded at 30fps), and the
            smallest possible frame-to-frame time-bin in Video 2 is 66ms (recorded at 15fps) we have to correct for
            this across recordings using the ``bin_size`` argument. E.g., when passing angular data from Video 1
            we would set bin_size to ``2``, and when passing angular data for Video 2 we would set bin_size to ``1`` to
            allow comparisons of instantaneous angular velocity between Video 1 and Video 2.

            When current frame minus bin_size results in a negative index, 0 is returned.

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter int bin_size: The number of frames prior to compare the current angular velocity against.

        :example:
        >>> data = np.array([350, 355, 356, 357])
        >>> FeatureExtractionCircularMixin().instantaneous_angular_velocity(data=data, bin_size=1)
        >>> [0., 5., 1., 1.]
        >>> FeatureExtractionCircularMixin().instantaneous_angular_velocity(data=data, bin_size=2)
        >>> [0., 0., 6., 2.]
        """
        data = np.deg2rad(data)
        results = np.full((data.shape[0]), 0.0)
        left_idx, right_idx = 0, bin_size
        for end_idx in prange(right_idx, data.shape[0] + 1, 1):
            results[end_idx] = np.rad2deg(
                np.pi - np.abs(np.pi - np.abs(data[left_idx] - data[end_idx]))
            )
            left_idx += 1
        return results

    @staticmethod
    def sliding_rao_spacing(
        data: np.ndarray, time_window: float, fps: int
    ) -> np.ndarray:
        """
        Compute the uniformity of a circular dataset in sliding window of size ``time_window``.

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Rolling time-window as float in seconds.
        :parameter int fps: fps of the recorded video
        :return np.ndarray: representing rao-spacing U in every sliding windows [-window:n]

        .. image:: _static/img/raospacing.png
           :width: 800
           :align: center

        :references:
        .. [1] `UCSB <https://jammalam.faculty.pstat.ucsb.edu/html/favorite/test.htm>`__.

        :example:
        >>> data = np.random.randint(low=0, high=360, size=(500,))
        >>> result = FeatureExtractionCircularMixin().sliding_rao_spacing(data=data, time_window=0.5, fps=10)
        """

        results = np.full((data.shape[0]), -1.0)
        window_size = int(time_window * fps)
        for i in range(window_size, data.shape[0]):
            w_data = np.sort(data[i - window_size : i])
            Ti, TiL = np.full((w_data.shape[0]), np.nan), np.full(
                (w_data.shape[0]), np.nan
            )
            l = 360 / len(w_data)
            Ti[-1] = np.rad2deg(
                np.pi
                - np.abs(np.pi - np.abs(np.deg2rad(w_data[0]) - np.deg2rad(w_data[-1])))
            )
            for j in prange(w_data.shape[0] - 1, -1, -1):
                Ti[j] = np.rad2deg(
                    np.pi
                    - np.abs(
                        np.pi
                        - np.abs(np.deg2rad(w_data[j]) - np.deg2rad(w_data[j - 1]))
                    )
                )
            for k in prange(Ti.shape[0]):
                TiL[k] = np.max((l, Ti[k])) - np.min((l, Ti[k]))
            S = np.sum(TiL)
            U = int(S / 2)
            results[i] = U
        return results


# data = np.random.normal(loc=90, scale=360, size=500)
# data = np.random.randint(low=0, high=360, size=(500,))
# result = FeatureExtractionCircularMixin().sliding_rao_spacing(data=data, time_window=0.5, fps=10)
# print(result)
# raospacing(np.deg2rad(data[-6:-1]), axis=0)

# #data = np.random.normal(loc=90, scale=2, size=20)
# data = np.array([350, 355, 356, 357])
# angle_data = FeatureExtractionCircularMixin().instantaneous_angular_velocity(data=data, bin_size=2)
#
#


# data = np.random.normal(loc=45, scale=3, size=20)
# FeatureExtractionCircularMixin().agg_angular_difference_timebins(data=data,time_windows=np.array([1]), fps=5)


# nose_loc = np.random.randint(low=0, high=500, size=(200, 2)).astype('float32')
# left_ear_loc = np.random.randint(low=0, high=500, size=(200, 2)).astype('float32')
#
# angle_data = FeatureExtractionCircularMixin().direction_two_bps(bp_x=nose_loc, bp_y=left_ear_loc)
#


# data_1 = np.random.normal(loc=45, scale=3, size=20)
# data_2 = np.random.normal(loc=45, scale=150, size=30)
# data = np.hstack([data_1, data_2])
# FeatureExtractionCircularMixin().rolling_resultant_vector_length(data=data,time_windows=np.array([1]), fps=5)
#
#


# data = np.random.normal(loc=45, scale=1, size=20)
# FeatureExtractionCircularMixin().rolling_mean_dispersion(data=data,time_windows=np.array([0.5]), fps=10)


# data = np.array(list(range(0, 405, 45)))
# results = FeatureExtractionCircularMixin().degrees_to_compass_cardinal(degree_angles=data)


data = np.array(list(range(0, 405, 45)))
results = FeatureExtractionCircularMixin().degrees_to_compass_cardinal(
    degree_angles=data
)


# def direction_two_bps(bp_x: np.ndarray,
#                       bp_y: np.ndarray) -> np.ndarray:


# right_ear_loc = np.random.randint(low=0, high=500, size=(200, 2)).astype('float32')
# angle_data = FeatureExtractionCircularMixin().head_direction(nose_loc=nose_loc, left_ear_loc=left_ear_loc, right_ear_loc=right_ear_loc)
#
# resultant_length = FeatureExtractionCircularMixin().rolling_resultant_vector_length(data=angle_data.astype(np.int8), time_window=1.0, fps=25)
#
# #resultant_length = FeatureExtractionCircularMixin().rolling_rayleigh_z(data=angle_data.astype(np.int8), time_window=2.0, fps=5)

# start = time.time()
# correlation = FeatureExtractionCircularMixin().rolling_circular_correlation(data_x=angle_data.astype(np.int8), data_y=angle_data.astype(np.int8), time_window=2.0, fps=5)
# print(time.time() - start)



__author__ = "Simon Nilsson"

import glob
import os
from collections import defaultdict
from typing import List

import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_if_filepath_list_is_empty, check_int
from simba.utils.data import detect_bouts
from simba.utils.enums import Options
from simba.utils.errors import InvalidInputError, NoChoosenMeasurementError
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class TimeBinsClfCalculator(ConfigReader):
    """
    Computes aggregate classification results in user-defined time-bins. Results are stored in
    the ``project_folder/logs`` directory of the SimBA project`

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter int bin_length: Integer representing the time bin size in seconds
    :parameter List[str] measurements: Aggregate statistic measures calculated for each time bin. OPTIONS: ['First occurrence (s)', 'Event count',
        Total event duration (s)', 'Mean event duration (s)', 'Median event duration (s)', 'Mean event interval (s)',
        'Median event interval (s)']
    :parameter List[str] classifiers: Names of classifiers to calculate aggregate statistics in time-bins for. EXAMPLE: ['Attack', 'Sniffing']

    .. note::
    `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#part-4--analyze-machine-results>`__.

    Example
    ----------
    >>> timebin_clf_analyzer = TimeBinsClfCalculator(config_path='MyConfigPath', bin_length=15, measurements=['Event count', 'Total event duration (s)'])
    >>> timebin_clf_analyzer.run()

    """

    def __init__(
        self,
        config_path: str,
        bin_length: int,
        measurements: List[
            Literal[
                "First occurrence (s)",
                "Event count",
                "Total event duration (s)",
                "Mean event duration (s)",
                "Median event duration (s)",
                "Mean event interval (s)",
                "Median event interval (s)",
            ]
        ],
        classifiers: List[str],
    ):
        super().__init__(config_path=config_path)
        if len(measurements) == 0:
            raise NoChoosenMeasurementError()
        if (
            len(
                set(measurements).intersection(
                    set(Options.TIMEBINS_MEASURMENT_OPTIONS.value)
                )
            )
            == 0
        ):
            raise InvalidInputError(
                msg=f"No valid input data measures. OPTIONS: {Options.TIMEBINS_MEASURMENT_OPTIONS.value}"
            )
        check_int(name="Bin length", value=bin_length, min_value=1)
        self.bin_length, self.measurements, self.classifiers = (
            int(bin_length),
            measurements,
            classifiers,
        )
        self.files_found = glob.glob(self.machine_results_dir + "/*." + self.file_type)
        check_if_filepath_list_is_empty(
            filepaths=self.files_found,
            error_msg=f"SIMBA ERROR: Cannot perform time-bin classification analysis, no data in {self.machine_results_dir} directory",
        )
        print("Processing {} video(s)...".format(str(len(self.files_found))))
        self.out_df_lst = []

    def run(self):
        """
        Method for running the classifier time-bin analysis. Results are stored in the
        ``project_folder/logs`` directory of the SimBA project.

        Returns
        ----------
        None
        """

        video_dict = {}
        for file_cnt, file_path in enumerate(self.files_found):
            dir_name, file_name, extension = get_fn_ext(file_path)
            data_df = read_df(file_path, self.file_type)
            video_settings, px_per_mm, fps = self.read_video_info(video_name=file_name)
            fps = int(fps)
            bin_frame_length = self.bin_length * fps
            data_df_lst = [
                data_df[i : i + bin_frame_length]
                for i in range(0, data_df.shape[0], bin_frame_length)
            ]
            video_dict[file_name] = {}
            for bin_cnt, df in enumerate(data_df_lst):
                video_dict[file_name][bin_cnt] = {}
                bouts_df = detect_bouts(
                    data_df=df, target_lst=list(self.clf_names), fps=fps
                )
                bouts_df["Shifted start"] = bouts_df["Start_time"].shift(-1)
                bouts_df["Interval duration"] = (
                    bouts_df["Shifted start"] - bouts_df["End Time"]
                )
                for clf in self.clf_names:
                    video_dict[file_name][bin_cnt][clf] = defaultdict(list)
                    bout_df = bouts_df.loc[bouts_df["Event"] == clf]
                    if len(bouts_df) > 0:
                        video_dict[file_name][bin_cnt][clf]["First occurrence (s)"] = (
                            round(bout_df["Start_time"].min(), 3)
                        )
                        video_dict[file_name][bin_cnt][clf]["Event count"] = len(
                            bout_df
                        )
                        video_dict[file_name][bin_cnt][clf][
                            "Total event duration (s)"
                        ] = round(bout_df["Bout_time"].sum(), 3)
                        video_dict[file_name][bin_cnt][clf][
                            "Mean event duration (s)"
                        ] = round(bout_df["Bout_time"].mean(), 3)
                        video_dict[file_name][bin_cnt][clf][
                            "Median event duration (s)"
                        ] = round(bout_df["Bout_time"].median(), 3)
                    else:
                        video_dict[file_name][bin_cnt][clf][
                            "First occurrence (s)"
                        ] = None
                        video_dict[file_name][bin_cnt][clf]["Event count"] = 0
                        video_dict[file_name][bin_cnt][clf][
                            "Total event duration (s)"
                        ] = 0
                        video_dict[file_name][bin_cnt][clf][
                            "Mean event duration (s)"
                        ] = 0
                        video_dict[file_name][bin_cnt][clf][
                            "Median event duration (s)"
                        ] = 0
                    if len(bouts_df) > 1:
                        video_dict[file_name][bin_cnt][clf][
                            "Mean event interval (s)"
                        ] = round(bout_df[:-1]["Interval duration"].mean(), 3)
                        video_dict[file_name][bin_cnt][clf][
                            "Median event interval (s)"
                        ] = round(bout_df[:-1]["Interval duration"].median(), 3)
                    else:
                        video_dict[file_name][bin_cnt][clf][
                            "Mean event interval (s)"
                        ] = None
                        video_dict[file_name][bin_cnt][clf][
                            "Median event interval (s)"
                        ] = None

        for video_name, video_info in video_dict.items():
            for bin_number, bin_data in video_info.items():
                data_df = (
                    pd.DataFrame.from_dict(bin_data)
                    .reset_index()
                    .rename(columns={"index": "Measurement"})
                )
                data_df = pd.melt(data_df, id_vars=["Measurement"]).rename(
                    columns={"value": "Value", "variable": "Classifier"}
                )
                data_df.insert(loc=0, column="Time bin #", value=bin_number)
                data_df.insert(loc=0, column="Video", value=video_name)
                self.out_df_lst.append(data_df)
        out_df = (
            pd.concat(self.out_df_lst, axis=0)
            .sort_values(by=["Video", "Time bin #"])
            .set_index("Video")
        )
        out_df = out_df[out_df["Measurement"].isin(self.measurements)]
        out_df = out_df[out_df["Classifier"].isin(self.classifiers)]
        self.save_path = os.path.join(
            self.project_path, "logs", f"Time_bins_ML_results_{self.datetime}.csv"
        )
        out_df.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f'Classification time-bins results saved at project_folder/logs/output/{str("Time_bins_ML_results_" + self.datetime + ".csv")}',
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = TimeBinsClf(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                    bin_length=2,
#                    measurements=['First occurrence (s)', 'Event count', 'Total event duration (s)', 'Mean event duration (s)'],
#                    classifiers=['Attack', 'Sniffing'])
# test.analyze_timebins_clf()


# test = TimeBinsClf(config_path='/Users/simon/Desktop/troubleshooting/light_analyzer/project_folder/project_config.ini',
#                    bin_length=2,
#                    measurements=['First occurrence (s)', 'Event count', 'Total event duration (s)', 'Mean event duration (s)'])
# test.analyze_timebins_clf()



import os
from tkinter import messagebox
from typing import Union

from simba.mixins.config_reader import ConfigReader

try:
    from typing import Literal
except:
    from typing_extensions import Literal


class RemoveROIFeaturesPopUp:
    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        dataset: Literal["features_extracted", "targets_inserted"],
    ):
        answer = messagebox.askyesno(
            title="REMOVE ROI FEATURES",
            message=f"Are tou sure you want to remove ROI \nfeatures from data in the \nproject_folder/csv/{dataset} folder?",
        )

        if answer:
            config = ConfigReader(config_path=config_path)
            config.read_roi_data()
            if dataset == "targets_inserted":
                config.remove_roi_features(config.targets_folder)
            if dataset == "features_extracted":
                config.remove_roi_features(config.features_dir)
        else:
            pass


# test = RemoveROIFeaturesPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Nastacia_unsupervised/project_folder/csv/features_extracted',
#                          dataset='features_extracted')



__author__ = "Simon Nilsson"

import functools
import multiprocessing
import os
import platform
from typing import Dict, List

import numpy as np
from numba import jit

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    read_df)


class DistancePlotterMultiCore(ConfigReader, PlottingMixin):
    """
     Visualize the distances between pose-estimated body-parts (e.g., two animals) through line
     charts. Results are saved as individual line charts, and/or a video of line charts.
     Uses multiprocessing.

     :param str config_path: path to SimBA project config file in Configparser format
     :param bool frame_setting: If True, creates individual frames.
     :param bool video_setting: If True, creates videos.
     :param bool final_img: If True, creates a single .png representing the entire video.
     :param dict style_attr: Video style attributes (font sizes, line opacity etc.)
     :param dict files_found: Files to visualize.
     :param dict line_attr: Representing the body-parts to visualize the distance between and their colors.

    .. note::
       `GitHub tutorial/documentation <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

    Examples
    -----
    >>> style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8, 'opacity': 0.5}
    >>> line_attr = {0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}
    >>> distance_plotter = DistancePlotterMultiCore(config_path=r'/tests_/project_folder/project_config.ini', frame_setting=False, video_setting=True, final_img=True, style_attr=style_attr, line_attr=line_attr,  files_found=['/test_/project_folder/csv/machine_results/Together_1.csv'], core_cnt=5)
    >>> distance_plotter.run()
    """

    def __init__(
        self,
        config_path: str,
        frame_setting: bool,
        video_setting: bool,
        final_img: bool,
        files_found: List[str],
        style_attr: Dict[str, int],
        line_attr: Dict[int, list],
        core_cnt: int,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)

        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)

        (
            self.video_setting,
            self.frame_setting,
            self.files_found,
            self.style_attr,
            self.line_attr,
            self.final_img,
        ) = (
            video_setting,
            frame_setting,
            files_found,
            style_attr,
            line_attr,
            final_img,
        )
        if (not frame_setting) and (not video_setting) and (not self.final_img):
            raise NoSpecifiedOutputError(
                msg="Please choice to create frames and/or video distance plots"
            )
        self.core_cnt = core_cnt
        if not os.path.exists(self.line_plot_dir):
            os.makedirs(self.line_plot_dir)
        check_if_filepath_list_is_empty(
            filepaths=self.outlier_corrected_dir,
            error_msg="SIMBA ERROR: Zero files found in the project_folder/csv/machine_results directory. Create classification results before visualizing distances.",
        )
        print(f"Processing {str(len(self.files_found))} videos...")

    @staticmethod
    @jit(nopython=True)
    def __insert_group_idx_column(data: np.array, group: int):
        group_col = np.full((data.shape[0], 1), group)
        return np.hstack((group_col, data))

    def run(self):
        """
        Creates line charts. Results are stored in the `project_folder/frames/line_plot` directory

        Returns
        ----------
        None
        """

        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            self.data_df = read_df(file_path, self.file_type)
            distance_arr = np.full(
                (len(self.data_df), len(self.line_attr.keys())), np.nan
            )
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_info, self.px_per_mm, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            self.save_video_folder = os.path.join(self.line_plot_dir, self.video_name)
            self.temp_folder = os.path.join(self.line_plot_dir, self.video_name, "temp")
            self.save_frame_folder_dir = os.path.join(
                self.line_plot_dir, self.video_name
            )
            for distance_cnt, data in enumerate(self.line_attr.values()):
                distance_arr[:, distance_cnt] = (
                    np.sqrt(
                        (self.data_df[data[0] + "_x"] - self.data_df[data[1] + "_x"])
                        ** 2
                        + (self.data_df[data[0] + "_y"] - self.data_df[data[1] + "_y"])
                        ** 2
                    )
                    / self.px_per_mm
                ) / 10
            if self.frame_setting:
                if os.path.exists(self.save_frame_folder_dir):
                    self.remove_a_folder(self.save_frame_folder_dir)
                if not os.path.exists(self.save_frame_folder_dir):
                    os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                self.video_folder = os.path.join(self.line_plot_dir, self.video_name)
                if os.path.exists(self.temp_folder):
                    self.remove_a_folder(self.temp_folder)
                    self.remove_a_folder(self.video_folder)
                os.makedirs(self.temp_folder)
                self.save_video_path = os.path.join(
                    self.line_plot_dir, self.video_name + ".mp4"
                )

            distance_arr = np.nan_to_num(distance_arr, nan=0.0)

            if self.final_img:
                self.make_distance_plot(
                    data=distance_arr,
                    line_attr=self.line_attr,
                    style_attr=self.style_attr,
                    fps=self.fps,
                    save_path=os.path.join(
                        self.line_plot_dir, self.video_name + "_final_img.png"
                    ),
                )

            if self.video_setting or self.frame_setting:
                if self.style_attr["y_max"] == "auto":
                    self.style_attr["max_y"] = np.amax(distance_arr)
                else:
                    self.style_attr["max_y"] = float(self.style_attr["max_y"])
                self.style_attr["y_ticks_locs"] = np.round(
                    np.linspace(0, self.style_attr["max_y"], 10), 2
                )
                self.style_attr["y_ticks_lbls"] = np.round(
                    (self.style_attr["y_ticks_locs"] / self.fps), 1
                )
                index_column = list(range(0, distance_arr.shape[0]))
                distance_arr = np.column_stack((index_column, distance_arr))

                distance_arr = np.array_split(distance_arr, self.core_cnt)

                data = []
                for cnt, i in enumerate(distance_arr):
                    data.append(self.__insert_group_idx_column(data=i, group=cnt))
                frm_per_core = data[0].shape[0]

                print(
                    "Creating distance plots, multiprocessing (chunksize: {}, cores: {})...".format(
                        str(self.multiprocess_chunksize), str(self.core_cnt)
                    )
                )
                with multiprocessing.Pool(
                    self.core_cnt, maxtasksperchild=self.maxtasksperchild
                ) as pool:
                    constants = functools.partial(
                        self.distance_plotter_mp,
                        video_setting=self.video_setting,
                        video_name=self.video_name,
                        frame_setting=self.frame_setting,
                        video_save_dir=self.temp_folder,
                        frame_folder_dir=self.save_frame_folder_dir,
                        style_attr=self.style_attr,
                        line_attr=self.line_attr,
                        fps=self.fps,
                    )
                    for cnt, result in enumerate(
                        pool.imap(
                            constants, data, chunksize=self.multiprocess_chunksize
                        )
                    ):
                        print(
                            "Image {}/{}, Video {}/{}...".format(
                                str(int(frm_per_core * (result + 1))),
                                str(len(self.data_df)),
                                str(file_cnt + 1),
                                str(len(self.files_found)),
                            )
                        )

                    pool.terminate()
                    pool.join()

                if self.video_setting:
                    print("Joining {} multiprocessed video...".format(self.video_name))
                    concatenate_videos_in_folder(
                        in_folder=self.temp_folder, save_path=self.save_video_path
                    )

                video_timer.stop_timer()
                print(
                    "Distance line chart video {} complete (elapsed time: {}s) ...".format(
                        self.video_name, video_timer.elapsed_time_str
                    )
                )

        self.timer.stop_timer()
        stdout_success(
            f"Distance plot visualizations for {str(len(self.files_found))} video(s) created in project_folder/frames/output/line_plot directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8, 'y_max': 'auto', 'opacity': 0.9}
# line_attr = {0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}
#
# test = DistancePlotterMultiCore(config_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 frame_setting=False,
#                                 video_setting=True,
#                                 style_attr=style_attr,
#                                 final_img=True,
#                                 files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                                 line_attr=line_attr,
#                                 core_cnt=3)
# test.create_distance_plot()
# #
# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8}
# line_attr = {0: ['Termite_1_Head_1', 'Termite_1_Thorax_1', 'Dark-red']}
#


# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8, 'opacity': 0.5, 'y_max': 'auto'}
# line_attr = {0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}
#
# test = DistancePlotterMultiCore(config_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                        frame_setting=False,
#                        video_setting=True,
#                        style_attr=style_attr,
#                                  final_img=False,
#                        files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                        line_attr=line_attr,
#                                 core_cnt=5)
# test.create_distance_plot()

# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8}
# line_attr = {0: ['Termite_1_Head_1', 'Termite_1_Thorax_1', 'Dark-red']}

# test = DistancePlotterSingleCore(config_path=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini',
#                                  frame_setting=False,
#                        video_setting=True,
#                        style_attr=style_attr,
#                        files_found=['/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/csv/outlier_corrected_movement_location/termites_1.csv'],
#                        line_attr=line_attr)
# test.create_distance_plot()
